'''
'''
#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import time

Test.Summary = '''
cache_range_requests cache-complete-responses test
'''

# Test description:
# Three rounds of testing:
#  Round 1:
#   - seed the cache with an object that is smaller than the slice block size
#   - issue requests with various ranges and validate responses are 200s
#  Round 2:
#   - seed the cache with an object that is larger than the slice block size
#   - issue requests with various ranges and validate responses are 206s
# The first two rounds test cache miss, cache hit, and refresh hit scenarios
#   - uses the cachekey plugin to add the `Range` request header to the cache key
#   - requests content through the slice and cache_range_requests plugin with a 4MB slice block size
#   - demonstrates how one might normalize the `Range` header to avoid cache pollution
# The third round tests cache miss, hit, and refresh hit scenarios without any other plugins
#   - tests cache misses, then hits, using the same object but two different ranges
#   - demonstrates why normalization of the `Range` header is required to prevent pollution

Test.SkipUnless(
    Condition.PluginExists('cachekey.so'),
    Condition.PluginExists('cache_range_requests.so'),
    Condition.PluginExists('slice.so'),
    Condition.PluginExists('xdebug.so'),
)
Test.ContinueOnFail = False
Test.testName = "cache_range_requests_cache_200s"

# Generate bodies for our responses
small_body_len = 10000
small_body = 'x' * small_body_len

slice_body_len = 4 * 1024 * 1024
slice_body = 'x' * slice_body_len

# Define and configure ATS
ts = Test.MakeATSProcess("ts")

# Define and configure origin server
server = Test.MakeOriginServer("server", lookup_key="{%UID}")

# default root
req_chk = {
    "headers": "GET / HTTP/1.1\r\n" + "Host: www.example.com\r\n" + "uuid: none\r\n" + "\r\n",
    "timestamp": "1469733493.993",
    "body": ""
}

res_chk = {"headers": "HTTP/1.1 200 OK\r\n" + "Connection: close\r\n" + "\r\n", "timestamp": "1469733493.993", "body": ""}

server.addResponse("sessionlog.json", req_chk, res_chk)

small_req = {
    "headers": "GET /obj HTTP/1.1\r\n" + "Host: www.example.com\r\n" + "Accept: */*\r\n" + "UID: SMALL\r\n"
               "\r\n",
    "timestamp": "1469733493.993",
    "body": ""
}

small_resp = {
    "headers":
        "HTTP/1.1 200 OK\r\n" + "Cache-Control: max-age=1\r\n" + "Connection: close\r\n" + 'Etag: "772102f4-56f4bc1e6d417"\r\n' +
        "\r\n",
    "timestamp": "1469733493.993",
    "body": small_body
}

small_reval_req = {
    "headers": "GET /obj HTTP/1.1\r\n" + "Host: www.example.com\r\n" + "Accept: */*\r\n" + "UID: SMALL-INM\r\n"
               "\r\n",
    "timestamp": "1469733493.993",
    "body": ""
}

small_reval_resp = {
    "headers":
        "HTTP/1.1 304 Not Modified\r\n" + "Cache-Control: max-age=10\r\n" + "Connection: close\r\n" +
        'Etag: "772102f4-56f4bc1e6d417"\r\n' + "\r\n",
    "timestamp": "1469733493.993"
}

slice_req = {
    "headers":
        "GET /slice HTTP/1.1\r\n" + "Host: www.example.com\r\n" + "Range: bytes=0-4194303\r\n" + "Accept: */*\r\n" +
        "UID: SLICE\r\n"
        "\r\n",
    "timestamp": "1469733493.993",
}

slice_resp = {
    "headers":
        "HTTP/1.1 206 Partial Content\r\n" + "Cache-Control: max-age=1\r\n" +
        "Content-Range: bytes 0-{}/{}\r\n".format(slice_body_len - 1, slice_body_len * 2) + "\r\n" +
        "Content-Length: {}\r\n".format(slice_body_len) + "\r\n" + "Connection: close\r\n" + 'Etag: "872104f4-d6bcaa1e6f979"\r\n' +
        "\r\n",
    "timestamp": "1469733493.993",
    "body": slice_body
}

slice_reval_req = {
    "headers": "GET /slice HTTP/1.1\r\n" + "Host: www.example.com\r\n" + "Accept: */*\r\n" + "UID: SLICE-INM\r\n"
               "\r\n",
    "timestamp": "1469733493.993",
    "body": ""
}

slice_reval_resp = {
    "headers":
        "HTTP/1.1 304 Not Modified\r\n" + "Cache-Control: max-age=10\r\n" + "Connection: close\r\n" +
        'Etag: "872104f4-d6bcaa1e6f979"\r\n' + "\r\n",
    "timestamp": "1469733493.993"
}

naieve_req = {
    "headers": "GET /naieve/obj HTTP/1.1\r\n" + "Host: www.example.com\r\n" + "Accept: */*\r\n" + "UID: NAIEVE\r\n"
               "\r\n",
    "timestamp": "1469733493.993",
    "body": ""
}

naieve_resp = {
    "headers":
        "HTTP/1.1 200 OK\r\n" + "Cache-Control: max-age=1\r\n" + "Connection: close\r\n" + 'Etag: "cad04ff4-56f4bc197ceda"\r\n' +
        "\r\n",
    "timestamp": "1469733493.993",
    "body": small_body
}

naieve_reval_req = {
    "headers": "GET /naieve/obj HTTP/1.1\r\n" + "Host: www.example.com\r\n" + "Accept: */*\r\n" + "UID: NAIEVE-INM\r\n"
               "\r\n",
    "timestamp": "1469733493.993",
    "body": ""
}

naieve_reval_resp = {
    "headers":
        "HTTP/1.1 304 Not Modified\r\n" + "Cache-Control: max-age=10\r\n" + "Connection: close\r\n" +
        'Etag: "cad04ff4-56f4bc197ceda"\r\n' + "\r\n",
    "timestamp": "1469733493.993"
}

server.addResponse("sessionlog.json", small_req, small_resp)
server.addResponse("sessionlog.json", small_reval_req, small_reval_resp)
server.addResponse("sessionlog.json", slice_req, slice_resp)
server.addResponse("sessionlog.json", slice_reval_req, slice_reval_resp)
server.addResponse("sessionlog.json", naieve_req, naieve_resp)
server.addResponse("sessionlog.json", naieve_reval_req, naieve_reval_resp)

# remap with the cache range requests plugin only
# this is a "naieve" configuration due to the lack of range normalization performed at remap time by slice
# this config should only be used if ranges have been reliably normalized by the requestor (either the client itself or a cache)
ts.Disk.remap_config.AddLines(
    [
        f'map http://example.com/naieve http://127.0.0.1:{server.Variables.Port}/naieve \\' +
        ' @plugin=cache_range_requests.so @pparam=--cache-complete-responses',
    ])

# remap with slice, cachekey, and the cache range requests plugin to ensure range normalization and cache keys are correct
ts.Disk.remap_config.AddLines(
    [
        f'map http://example.com http://127.0.0.1:{server.Variables.Port} \\' + ' @plugin=slice.so @pparam=--blockbytes=4m \\',
        ' @plugin=cachekey.so @pparam=--key-type=cache_key @pparam=--include-headers=Range @pparam=--remove-all-params=true \\',
        ' @plugin=cache_range_requests.so @pparam=--no-modify-cachekey @pparam=--cache-complete-responses',
    ])

# cache debug
ts.Disk.plugin_config.AddLine('xdebug.so --enable=x-cache,x-cache-key')

# enable debug
ts.Disk.records_config.update(
    {
        'proxy.config.diags.debug.enabled': 1,
        'proxy.config.diags.debug.tags': 'cachekey|cache_range_requests|slice',
    })

# base cURL command
curl_and_args = 'curl -s -D /dev/stdout -o /dev/stderr -x localhost:{} -H "x-debug: x-cache, x-cache-key"'.format(ts.Variables.port)

# Test round 1: ensure we fetch and cache objects that are returned with a
# 200 OK and no Content-Range when the object is smaller than the slice
# block size

# 0 Test - Fetch /obj with a Range header but less than 4MB
tr = Test.AddTestRun("cache miss on /obj")
ps = tr.Processes.Default
ps.StartBefore(server, ready=When.PortOpen(server.Variables.Port))
ps.StartBefore(Test.Processes.ts)
ps.Command = curl_and_args + ' -H "UID: SMALL" http://example.com/obj -r 0-5000'
ps.ReturnCode = 0
ps.Streams.stdout.Content = Testers.ContainsExpression("200 OK", "expected 200 OK")
ps.Streams.stdout.Content = Testers.ExcludesExpression("Content-Range:", "expected no Content-Range header")
ps.Streams.stdout.Content = Testers.ContainsExpression("X-Cache: miss, none", "expected cache miss")
ps.Streams.stdout.Content = Testers.ContainsExpression(
    "X-Cache-Key: /.*?/Range:bytes=0-4194303/obj", "expected cache key with bytes 0-4194303")
tr.StillRunningAfter = ts

# 1 Test - Fetch /obj with a different range but less than 4MB
tr = Test.AddTestRun("cache hit-fresh on /obj")
ps = tr.Processes.Default
ps.Command = curl_and_args + ' -H "UID: SMALL" http://example.com/obj -r 5001-5999'
ps.ReturnCode = 0
ps.Streams.stdout.Content = Testers.ContainsExpression("200 OK", "expected 200 OK")
ps.Streams.stdout.Content = Testers.ExcludesExpression("Content-Range:", "expected no Content-Range header")
ps.Streams.stdout.Content = Testers.ContainsExpression("X-Cache: hit-fresh, none", "expected cache hit")
ps.Streams.stdout.Content = Testers.ContainsExpression(
    "X-Cache-Key: /.*?/Range:bytes=0-4194303/obj", "expected cache key with bytes 0-4194303")
tr.StillRunningAfter = ts

# 2 Test - Revalidate /obj with a different range but less than 4MB
tr = Test.AddTestRun("cache hit-stale on /obj")
tr.DelayStart = 2
ps = tr.Processes.Default
ps.Command = curl_and_args + ' -H "UID: SMALL-INM" http://example.com/obj -r 0-403'
ps.ReturnCode = 0
ps.Streams.stdout.Content = Testers.ContainsExpression("200 OK", "expected 200 OK")
ps.Streams.stdout.Content = Testers.ExcludesExpression("Content-Range:", "expected no Content-Range header")
ps.Streams.stdout.Content = Testers.ContainsExpression("X-Cache: hit-stale, none", "expected cache hit stale")
ps.Streams.stdout.Content = Testers.ContainsExpression(
    "X-Cache-Key: /.*?/Range:bytes=0-4194303/obj", "expected cache key with bytes 0-4194303")
tr.StillRunningAfter = ts

# 3 Test - Fetch /obj with a different range but less than 4MB
tr = Test.AddTestRun("cache hit on /obj post revalidation")
ps = tr.Processes.Default
ps.Command = curl_and_args + ' -H "UID: SMALL" http://example.com/obj -r 0-3999'
ps.ReturnCode = 0
ps.Streams.stdout.Content = Testers.ContainsExpression("200 OK", "expected 200 OK")
ps.Streams.stdout.Content = Testers.ExcludesExpression("Content-Range:", "expected no Content-Range header")
ps.Streams.stdout.Content = Testers.ContainsExpression("X-Cache: hit-fresh, none", "expected cache hit-fresh")
ps.Streams.stdout.Content = Testers.ContainsExpression(
    "X-Cache-Key: /.*?/Range:bytes=0-4194303/obj", "expected cache key with bytes 0-4194303")
tr.StillRunningAfter = ts

# Test round 2: repeat, but ensure we have 206s and matching Content-Range
# headers due to a base object that exceeds the slice block size

# 4 Test - Fetch /slice with a Range header but less than 4MB
tr = Test.AddTestRun("cache miss on /slice")
ps = tr.Processes.Default
ps.Command = curl_and_args + ' -H "UID: SLICE" http://example.com/slice -r 0-5000'
ps.ReturnCode = 0
ps.Streams.stdout.Content = Testers.ContainsExpression("206 Partial Content", "expected 206 Partial Content")
ps.Streams.stdout.Content = Testers.ContainsExpression(
    "Content-Range: bytes 0-5000/8388608", "expected Content-Range: bytes 0-5000/8388608")
ps.Streams.stdout.Content = Testers.ContainsExpression("X-Cache: miss, none", "expected cache miss")
ps.Streams.stdout.Content = Testers.ContainsExpression(
    "X-Cache-Key: /.*?/Range:bytes=0-4194303/slice", "expected cache key with bytes 0-4194303")
tr.StillRunningAfter = ts

# 5 Test - Fetch /slice with a different range but less than 4MB
tr = Test.AddTestRun("cache hit-fresh on /slice")
ps = tr.Processes.Default
ps.Command = curl_and_args + ' -H "UID: SLICE" http://example.com/slice -r 5001-5999'
ps.ReturnCode = 0
ps.Streams.stdout.Content = Testers.ContainsExpression("206 Partial Content", "expected 206 Partial Content")
ps.Streams.stdout.Content = Testers.ContainsExpression(
    "Content-Range: bytes 5001-5999/8388608", "expected Content-Range: bytes 5001-5999/8388608")
ps.Streams.stdout.Content = Testers.ContainsExpression("X-Cache: hit-fresh, none", "expected cache hit")
ps.Streams.stdout.Content = Testers.ContainsExpression(
    "X-Cache-Key: /.*?/Range:bytes=0-4194303/slice", "expected cache key with bytes 0-4194303")
tr.StillRunningAfter = ts

# 6 Test - Revalidate /slice with a different range but less than 4MB
tr = Test.AddTestRun("cache hit-stale on /slice")
tr.DelayStart = 2
ps = tr.Processes.Default
ps.Command = curl_and_args + ' -H "UID: SLICE-INM" http://example.com/slice -r 0-403'
ps.ReturnCode = 0
ps.Streams.stdout.Content = Testers.ContainsExpression("206 Partial Content", "expected 206 Partial Content")
ps.Streams.stdout.Content = Testers.ContainsExpression(
    "Content-Range: bytes 0-403/8388608", "expected Content-Range: bytes 0-403/8388608")
ps.Streams.stdout.Content = Testers.ContainsExpression("X-Cache: hit-stale, none", "expected cache hit stale")
ps.Streams.stdout.Content = Testers.ContainsExpression(
    "X-Cache-Key: /.*?/Range:bytes=0-4194303/slice", "expected cache key with bytes 0-4194303")
tr.StillRunningAfter = ts

# 7 Test - Fetch /slice with a different range but less than 4MB
tr = Test.AddTestRun("cache hit on /slice post revalidation")
ps = tr.Processes.Default
ps.Command = curl_and_args + ' -H "UID: SLICE" http://example.com/slice -r 0-3999'
ps.ReturnCode = 0
ps.Streams.stdout.Content = Testers.ContainsExpression("206 Partial Content", "expected 206 Partial Content")
ps.Streams.stdout.Content = Testers.ContainsExpression(
    "Content-Range: bytes 0-3999/8388608", "expected Content-Range: bytes 0-3999/8388608")
ps.Streams.stdout.Content = Testers.ContainsExpression("X-Cache: hit-fresh, none", "expected cache hit-fresh")
ps.Streams.stdout.Content = Testers.ContainsExpression(
    "X-Cache-Key: /.*?/Range:bytes=0-4194303/slice", "expected cache key with bytes 0-4194303")
tr.StillRunningAfter = ts

# Test round 3: test behavior of the cache range requests plugin when caching complete ranges *without* the slice and cachekey plugins
# this tests the "naieve" case that requires range normalization to be
# performed by the requestor and demonstrates how the cache can be
# polluted without normalization

# 8 Test - Fetch /naieve/obj with a Range header
tr = Test.AddTestRun("cache miss on /naieve/obj")
ps = tr.Processes.Default
ps.Command = curl_and_args + ' -H "UID: NAIEVE" http://example.com/naieve/obj -r 0-5000'
ps.ReturnCode = 0
ps.Streams.stdout.Content = Testers.ContainsExpression("200 OK", "expected 200 OK")
ps.Streams.stdout.Content = Testers.ExcludesExpression("Content-Range:", "expected no Content-Range header")
ps.Streams.stdout.Content = Testers.ContainsExpression("X-Cache: miss", "expected cache miss")
ps.Streams.stdout.Content = Testers.ContainsExpression(
    "X-Cache-Key: http://.*?/naieve/obj-bytes=0-5000", "expected cache key with bytes 0-5000")
tr.StillRunningAfter = ts

# 9 Test - Fetch /naieve/obj with the same Range header
tr = Test.AddTestRun("cache hit-fresh on /naieve/obj")
ps = tr.Processes.Default
ps.Command = curl_and_args + ' -H "UID: NAIEVE" http://example.com/naieve/obj -r 0-5000'
ps.ReturnCode = 0
ps.Streams.stdout.Content = Testers.ContainsExpression("200 OK", "expected 200 OK")
ps.Streams.stdout.Content = Testers.ExcludesExpression("Content-Range:", "expected no Content-Range header")
ps.Streams.stdout.Content = Testers.ContainsExpression("X-Cache: hit-fresh", "expected cache hit")
ps.Streams.stdout.Content = Testers.ContainsExpression(
    "X-Cache-Key: http://.*?/naieve/obj-bytes=0-5000", "expected cache key with bytes 0-5000")
tr.StillRunningAfter = ts

# 10 Test - Revalidate /naieve/obj with the same Range header
tr = Test.AddTestRun("cache hit-stale on /naieve/obj")
tr.DelayStart = 2
ps = tr.Processes.Default
ps.Command = curl_and_args + ' -H "UID: NAIEVE-INM" http://example.com/naieve/obj -r 0-5000'
ps.ReturnCode = 0
ps.Streams.stdout.Content = Testers.ContainsExpression("200 OK", "expected 200 OK")
ps.Streams.stdout.Content = Testers.ExcludesExpression("Content-Range:", "expected no Content-Range header")
ps.Streams.stdout.Content = Testers.ContainsExpression("X-Cache: hit-stale", "expected cache hit stale")
ps.Streams.stdout.Content = Testers.ContainsExpression(
    "X-Cache-Key: http://.*?/naieve/obj-bytes=0-5000", "expected cache key with bytes 0-5000")
tr.StillRunningAfter = ts

# 11 Test - Fetch /naieve/obj with the same Range header
tr = Test.AddTestRun("cache hit on /naieve/obj post revalidation")
ps = tr.Processes.Default
ps.Command = curl_and_args + ' -H "UID: NAIEVE" http://example.com/naieve/obj -r 0-5000'
ps.ReturnCode = 0
ps.Streams.stdout.Content = Testers.ContainsExpression("200 OK", "expected 200 OK")
ps.Streams.stdout.Content = Testers.ExcludesExpression("Content-Range:", "expected no Content-Range header")
ps.Streams.stdout.Content = Testers.ContainsExpression("X-Cache: hit-fresh", "expected cache hit-fresh")
ps.Streams.stdout.Content = Testers.ContainsExpression(
    "X-Cache-Key: http://.*?/naieve/obj-bytes=0-5000", "expected cache key with bytes 0-5000")
tr.StillRunningAfter = ts

# 12 Test - Fetch /naieve/obj with a *different* Range header; note the cache key changes and is a miss for the same object
tr = Test.AddTestRun("cache miss on /naieve/obj")
ps = tr.Processes.Default
ps.Command = curl_and_args + ' -H "UID: NAIEVE" http://example.com/naieve/obj -r 444-777'
ps.ReturnCode = 0
ps.Streams.stdout.Content = Testers.ContainsExpression("200 OK", "expected 200 OK")
ps.Streams.stdout.Content = Testers.ExcludesExpression("Content-Range:", "expected no Content-Range header")
ps.Streams.stdout.Content = Testers.ContainsExpression("X-Cache: miss", "expected cache miss")
ps.Streams.stdout.Content = Testers.ContainsExpression(
    "X-Cache-Key: http://.*?/naieve/obj-bytes=444-777", "expected cache key with bytes 444-777")
tr.StillRunningAfter = ts

# 13 Test - Fetch /naieve/obj with the prior Range header; now a cache hit but we've effectively cached /naieve/obj twice
# this is why a Range normalization strategy should _always_ be employed when using `--cache-complete-responses`
tr = Test.AddTestRun("cache hit on /naieve/obj")
ps = tr.Processes.Default
ps.Command = curl_and_args + ' -H "UID: NAIEVE" http://example.com/naieve/obj -r 444-777'
ps.ReturnCode = 0
ps.Streams.stdout.Content = Testers.ContainsExpression("200 OK", "expected 200 OK")
ps.Streams.stdout.Content = Testers.ExcludesExpression("Content-Range:", "expected no Content-Range header")
ps.Streams.stdout.Content = Testers.ContainsExpression("X-Cache: hit", "expected cache hit-fresh")
ps.Streams.stdout.Content = Testers.ContainsExpression(
    "X-Cache-Key: http://.*?/naieve/obj-bytes=444-777", "expected cache key with bytes 444-777")
tr.StillRunningAfter = ts

# 14 Test - Fetch /naieve/obj with the original Range header (0-5000); still a cache hit
tr = Test.AddTestRun("cache hit on /naieve/obj after requesting with a different Range")
ps = tr.Processes.Default
ps.Command = curl_and_args + ' -H "UID: NAIEVE" http://example.com/naieve/obj -r 0-5000'
ps.ReturnCode = 0
ps.Streams.stdout.Content = Testers.ContainsExpression("200 OK", "expected 200 OK")
ps.Streams.stdout.Content = Testers.ExcludesExpression("Content-Range:", "expected no Content-Range header")
ps.Streams.stdout.Content = Testers.ContainsExpression("X-Cache: hit-fresh", "expected cache hit-fresh")
ps.Streams.stdout.Content = Testers.ContainsExpression(
    "X-Cache-Key: http://.*?/naieve/obj-bytes=0-5000", "expected cache key with bytes 0-5000")
tr.StillRunningAfter = ts



'''
Test offering client cert to origin
'''
#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

Test.Summary = '''
Test client certs to origin selected via wildcard names in sni
'''

ts = Test.MakeATSProcess("ts")
cafile = "{0}/signer.pem".format(Test.RunDirectory)
cafile2 = "{0}/signer2.pem".format(Test.RunDirectory)
server = Test.MakeOriginServer(
    "server",
    ssl=True,
    options={
        "--clientCA": cafile,
        "--clientverify": ""
    },
    clientcert="{0}/signed-foo.pem".format(Test.RunDirectory),
    clientkey="{0}/signed-foo.key".format(Test.RunDirectory))
server2 = Test.MakeOriginServer(
    "server2",
    ssl=True,
    options={
        "--clientCA": cafile2,
        "--clientverify": ""
    },
    clientcert="{0}/signed2-bar.pem".format(Test.RunDirectory),
    clientkey="{0}/signed-bar.key".format(Test.RunDirectory))
server4 = Test.MakeOriginServer("server4")
server.Setup.Copy("ssl/signer.pem")
server.Setup.Copy("ssl/signer2.pem")
server.Setup.Copy("ssl/signed-foo.pem")
server.Setup.Copy("ssl/signed-foo.key")
server.Setup.Copy("ssl/signed2-foo.pem")
server.Setup.Copy("ssl/signed2-bar.pem")
server.Setup.Copy("ssl/signed-bar.key")
server2.Setup.Copy("ssl/signer.pem")
server2.Setup.Copy("ssl/signer2.pem")
server2.Setup.Copy("ssl/signed-foo.pem")
server2.Setup.Copy("ssl/signed-foo.key")
server2.Setup.Copy("ssl/signed2-foo.pem")
server2.Setup.Copy("ssl/signed2-bar.pem")
server2.Setup.Copy("ssl/signed-bar.key")

request_header = {"headers": "GET / HTTP/1.1\r\nHost: example.com\r\n\r\n", "timestamp": "1469733493.993", "body": ""}
response_header = {"headers": "HTTP/1.1 200 OK\r\nConnection: close\r\n\r\n", "timestamp": "1469733493.993", "body": ""}
server.addResponse("sessionlog.json", request_header, response_header)
request_header = {"headers": "GET / HTTP/1.1\r\nHost: bar.com\r\n\r\n", "timestamp": "1469733493.993", "body": ""}
response_header = {"headers": "HTTP/1.1 200 OK\r\nConnection: close\r\n\r\n", "timestamp": "1469733493.993", "body": ""}
server.addResponse("sessionlog.json", request_header, response_header)

ts.addSSLfile("ssl/server.pem")
ts.addSSLfile("ssl/server.key")
ts.addSSLfile("ssl/combo-signed-foo.pem")
ts.addSSLfile("ssl/signed-foo.pem")
ts.addSSLfile("ssl/signed-foo.key")
ts.addSSLfile("ssl/signed2-foo.pem")
ts.addSSLfile("ssl/signed-bar.pem")
ts.addSSLfile("ssl/signed2-bar.pem")
ts.addSSLfile("ssl/signed-bar.key")

ts.Disk.records_config.update(
    {
        'proxy.config.ssl.server.cert.path': '{0}'.format(ts.Variables.SSLDir),
        'proxy.config.ssl.server.private_key.path': '{0}'.format(ts.Variables.SSLDir),
        'proxy.config.ssl.client.cert.path': '{0}'.format(ts.Variables.SSLDir),
        'proxy.config.ssl.client.private_key.path': '{0}'.format(ts.Variables.SSLDir),
        'proxy.config.exec_thread.autoconfig.scale': 1.0,
        'proxy.config.url_remap.pristine_host_hdr': 1,
        'proxy.config.ssl.client.verify.server.policy': 'PERMISSIVE',
    })

ts.Disk.ssl_multicert_config.AddLine('dest_ip=* ssl_cert_name=server.pem ssl_key_name=server.key')

ts.Disk.remap_config.AddLine('map /case1 https://127.0.0.1:{0}/'.format(server.Variables.SSL_Port))
ts.Disk.remap_config.AddLine('map /case2 https://127.0.0.1:{0}/'.format(server2.Variables.SSL_Port))

ts.Disk.sni_yaml.AddLines(
    [
        'sni:',
        '- fqdn: bob.bar.com',
        '  client_cert: signed-bar.pem',
        '  client_key: signed-bar.key',
        '- fqdn: "bob.foo.com"',
        '  client_cert: {0}/combo-signed-foo.pem'.format(ts.Variables.SSLDir),
        '- fqdn: "*.bar.com"',
        '  client_cert: {0}/signed2-bar.pem'.format(ts.Variables.SSLDir),
        '  client_key: {0}/signed-bar.key'.format(ts.Variables.SSLDir),
        '- fqdn: "foo.com"',
        '  client_cert: {0}/signed2-foo.pem'.format(ts.Variables.SSLDir),
        '  client_key: {0}/signed-foo.key'.format(ts.Variables.SSLDir),
    ])

ts.Disk.logging_yaml.AddLines(
    '''
logging:
  formats:
    - name: testformat
      format: '%<pssc> %<pquc> %<pscert> %<cscert>'
  logs:
    - mode: ascii
      format: testformat
      filename: squid
'''.split("\n"))

# Should succeed
tr = Test.AddTestRun("bob.bar.com to server 1")
tr.Processes.Default.StartBefore(Test.Processes.ts)
tr.Processes.Default.StartBefore(server)
tr.Processes.Default.StartBefore(server2)
tr.StillRunningAfter = ts
tr.StillRunningAfter = server
tr.StillRunningAfter = server2
tr.Processes.Default.Command = "curl -H host:bob.bar.com  http://127.0.0.1:{0}/case1".format(ts.Variables.port)
tr.Processes.Default.ReturnCode = 0
tr.Processes.Default.Streams.stdout = Testers.ExcludesExpression("Could Not Connect", "Check response")

# Should fail
trfail = Test.AddTestRun("bob.bar.com to server 2")
trfail.StillRunningAfter = ts
trfail.StillRunningAfter = server
trfail.StillRunningAfter = server2
trfail.Processes.Default.Command = 'curl -H host:bob.bar.com  http://127.0.0.1:{0}/case2'.format(ts.Variables.port)
trfail.Processes.Default.ReturnCode = 0
trfail.Processes.Default.Streams.stdout = Testers.ContainsExpression("Could Not Connect", "Check response")

# Should succeed
tr = Test.AddTestRun("bob.foo.com to server 1")
tr.StillRunningAfter = ts
tr.StillRunningAfter = server
tr.StillRunningAfter = server2
tr.Processes.Default.Command = "curl -H host:bob.foo.com  http://127.0.0.1:{0}/case1".format(ts.Variables.port)
tr.Processes.Default.ReturnCode = 0
tr.Processes.Default.Streams.stdout = Testers.ExcludesExpression("Could Not Connect", "Check response")

# Should fail
trfail = Test.AddTestRun("bob.foo.com to server 2")
trfail.StillRunningAfter = ts
trfail.StillRunningAfter = server
trfail.StillRunningAfter = server2
trfail.Processes.Default.Command = 'curl -H host:bob.foo.com  http://127.0.0.1:{0}/case2'.format(ts.Variables.port)
trfail.Processes.Default.ReturnCode = 0
trfail.Processes.Default.Streams.stdout = Testers.ContainsExpression("Could Not Connect", "Check response")

# Should succeed
tr = Test.AddTestRun("random.bar.com to server 2")
tr.StillRunningAfter = ts
tr.StillRunningAfter = server
tr.StillRunningAfter = server2
tr.Processes.Default.Command = "curl -H host:random.bar.com  http://127.0.0.1:{0}/case2".format(ts.Variables.port)
tr.Processes.Default.ReturnCode = 0
tr.Processes.Default.Streams.stdout = Testers.ExcludesExpression("Could Not Connect", "Check response")

# Should fail
trfail = Test.AddTestRun("random.bar.com to server 1")
trfail.StillRunningAfter = ts
trfail.StillRunningAfter = server
trfail.StillRunningAfter = server2
trfail.Processes.Default.Command = 'curl -H host:random.bar.com  http://127.0.0.1:{0}/case1'.format(ts.Variables.port)
trfail.Processes.Default.ReturnCode = 0
trfail.Processes.Default.Streams.stdout = Testers.ContainsExpression("Could Not Connect", "Check response")

# Should fail
tr = Test.AddTestRun("random.foo.com to server 2")
tr.StillRunningAfter = ts
tr.StillRunningAfter = server
tr.StillRunningAfter = server2
tr.Processes.Default.Command = "curl -H host:random.foo.com  http://127.0.0.1:{0}/case2".format(ts.Variables.port)
tr.Processes.Default.ReturnCode = 0
tr.Processes.Default.Streams.stdout = Testers.ContainsExpression("Could Not Connect", "Check response")

# Should fail
trfail = Test.AddTestRun("random.foo.com to server 1")
trfail.StillRunningAfter = ts
trfail.StillRunningAfter = server
trfail.StillRunningAfter = server2
trfail.Processes.Default.Command = 'curl -H host:random.foo.com  http://127.0.0.1:{0}/case1'.format(ts.Variables.port)
trfail.Processes.Default.ReturnCode = 0
trfail.Processes.Default.Streams.stdout = Testers.ContainsExpression("Could Not Connect", "Check response")

tr = Test.AddTestRun("Wait for the access log to write out")
tr.Processes.Default.StartBefore(server4, ready=When.FileExists(ts.Disk.squid_log))
tr.StillRunningAfter = ts
tr.Processes.Default.Command = 'echo "Log file exists"'
tr.Processes.Default.ReturnCode = 0

ts.Disk.squid_log.Content = "gold/proxycert2-accesslog.gold"



'''
'''
#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

Test.Summary = '''
Test a basic ip_resolve override using an ipv6 server
'''

Test.ContinueOnFail = True
# Define default ATS
ts = Test.MakeATSProcess("ts")
server = Test.MakeOriginServer("server")
server_v6 = Test.MakeOriginServer("server_v6", None, None, '::1', 0)

dns = Test.MakeDNServer("dns")

Test.testName = ""
request_header = {"headers": "GET / HTTP/1.1\r\nHost: www.example.com\r\n\r\n", "timestamp": "1469733493.993", "body": ""}
# expected response from the origin server
response_header = {"headers": "HTTP/1.1 200 OK\r\nConnection: close\r\n\r\n", "timestamp": "1469733493.993", "body": ""}

# add response to the server dictionary
server.addResponse("sessionfile.log", request_header, response_header)
server_v6.addResponse("sessionfile.log", request_header, response_header)
ts.Disk.records_config.update(
    {
        'proxy.config.diags.debug.enabled': 1,
        'proxy.config.diags.debug.tags': 'http.*|dns|conf_remap',
        'proxy.config.http.referer_filter': 1,
        'proxy.config.dns.nameservers': '127.0.0.1:{0}'.format(dns.Variables.Port),
        'proxy.config.dns.resolv_conf': 'NULL',
        'proxy.config.hostdb.ip_resolve': 'ipv4'
    })

ts.Disk.remap_config.AddLine(
    'map http://testDNS.com http://test.ipv4.only.com:{0}  @plugin=conf_remap.so @pparam=proxy.config.hostdb.ip_resolve=ipv6;ipv4;client'
    .format(server.Variables.Port))
ts.Disk.remap_config.AddLine(
    'map http://testDNS2.com http://test.ipv6.only.com:{0}  @plugin=conf_remap.so @pparam=proxy.config.hostdb.ip_resolve=ipv6;only'
    .format(server_v6.Variables.Port))

dns.addRecords(records={"test.ipv4.only.com.": ["127.0.0.1"]})
dns.addRecords(records={"test.ipv6.only.com": ["127.0.0.1", "::1"]})

tr = Test.AddTestRun()
tr.Processes.Default.Command = 'curl  --proxy 127.0.0.1:{0} "http://testDNS.com" --verbose'.format(ts.Variables.port)
tr.Processes.Default.ReturnCode = 0
tr.Processes.Default.StartBefore(server)
tr.Processes.Default.StartBefore(dns)
tr.Processes.Default.StartBefore(Test.Processes.ts)
tr.Processes.Default.Streams.stderr = "gold/remap-DNS-200.gold"
tr.StillRunningAfter = server

tr = Test.AddTestRun()
tr.Processes.Default.Command = 'curl  --proxy 127.0.0.1:{0} "http://testDNS2.com" --verbose'.format(ts.Variables.port)
tr.Processes.Default.ReturnCode = 0
tr.Processes.Default.StartBefore(server_v6)
tr.Processes.Default.Streams.stderr = "gold/remap-DNS-ipv6-200.gold"
tr.StillRunningAfter = server_v6



'''
'''
#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

Test.Summary = '''
Test requests with bad HTTP formats
'''

ArbitraryTimestamp = '12345678'

random_path = "/sdfsdf"

random_method = "xyzxyz"

ts = Test.MakeATSProcess("ts")

Test.GetTcpPort("upstream_port")

method_server = Test.Processes.Process(
    "method-server", "bash -c '" + Test.TestDirectory + f"/method-server.sh {Test.Variables.upstream_port} outserver'")

server = Test.MakeOriginServer("server", ssl=False)
request_header = {
    "headers": "GET {}/0 HTTP/1.1\r\nX-Req-Id: 0\r\nHost: example.com\r\n\r\n".format(random_path),
    "timestamp": "1469733493.993",
    "body": ""
}
response_header = {
    "headers": "HTTP/1.1 200 OK\r\nX-Resp-Id: 1\r\nConnection: close\r\n\r\n",
    "timestamp": "1469733493.993",
    "body": ""
}
server.addResponse("sessionlog.json", request_header, response_header)

request_header = {
    "headers": "GET {}/1 HTTP/1.1\r\nX-Req-Id: 1\r\nHost: example.com\r\n\r\n".format(random_path),
    "timestamp": "1469733493.993",
    "body": ""
}
response_header = {
    "headers": "HTTP/1.1 200 OK\r\nX-Resp-Id: 2\r\nConnection: close\r\n\r\n",
    "timestamp": "1469733493.993",
    "body": ""
}
server.addResponse("sessionlog.json", request_header, response_header)

request_header = {
    "headers": "GET /example/1 HTTP/1.1\r\nX-Req-Id: 6\r\nHost: example.com\r\n\r\n",
    "timestamp": "1469733493.993",
    "body": ""
}
response_header = {
    "headers": "HTTP/1.1 200 OK\r\nX-Resp-Id: 3\r\nConnection: close\r\n\r\n",
    "timestamp": "1469733493.993",
    "body": ""
}
server.addResponse("sessionlog.json", request_header, response_header)

request_header = {
    "headers": "GET /example/2 HTTP/1.1\r\nX-Req-Id: 7\r\nHost: example.com\r\n\r\n",
    "timestamp": "1469733493.993",
    "body": ""
}
response_header = {
    "headers": "HTTP/1.1 200 OK\r\nX-Resp-Id: 4\r\nConnection: close\r\n\r\n",
    "timestamp": "1469733493.993",
    "body": ""
}
server.addResponse("sessionlog.json", request_header, response_header)

ts.Disk.records_config.update(
    {
        'proxy.config.diags.debug.enabled': 1,
        'proxy.config.diags.debug.tags': 'http|dns',
        'proxy.config.url_remap.remap_required': 0,
        'proxy.config.http.strict_uri_parsing': 1,
        'proxy.config.http.cache.http': 0,
    })

# Allow random_method for 128.0.0.1, do not allow it for ::1.
#
ts.Disk.ip_allow_yaml.AddLines(
    [
        'ip_allow:',
        '  - apply: in',
        '    ip_addrs: 127.0.0.1',
        '    action: allow',
        '    methods:',
        '      - GET',
        f'      - {random_method}',
    ])
ts.Disk.ip_allow_yaml.AddLines([
    '  - apply: in',
    '    ip_addrs: ::1',
    '    action: allow',
    '    methods:',
    '      - GET',
])

ts.Disk.remap_config.AddLine('map /add-method http://127.0.0.1:{0}/'.format(Test.Variables.upstream_port))
ts.Disk.remap_config.AddLine('map / http://127.0.0.1:{0}/'.format(server.Variables.Port))

# Even if the request from the client is HTTP version 1.0, ATS's request to server will be HTTP version 1.1.
#
tr = Test.AddTestRun("success-1.0")
tr.Processes.Default.StartBefore(ts)
tr.Processes.Default.StartBefore(server)
tr.Processes.Default.Command = (
    f"printf 'GET {random_path}/0HTTP/1.0\r\n" + "Host: example.com\r\n" + "Connection: close\r\n" + "X-Req-Id: 1\r\n\r\n'" +
    f" | nc localhost {ts.Variables.port} >> client.log" + " ; echo '======' >> client.log")
tr.Processes.Default.ReturnCode = 0

tr = Test.AddTestRun("success-1.1")
tr.Processes.Default.Command = (
    f"printf 'GET {random_path}/1HTTP/1.1\r\n" + "Host: example.com\r\n" + "Connection: close\r\n" + "X-Req-Id: 2\r\n\r\n'" +
    f" | nc localhost {ts.Variables.port} >> client.log" + " ; echo '======' >> client.log")
tr.Processes.Default.ReturnCode = 0

tr = Test.AddTestRun("invalid-url-line")
tr.Processes.Default.Command = (
    f"printf 'GET {random_path}<HTTP/1.1\r\n" + "Host: example.com\r\n" + "Connection: close\r\n" + "X-Req-Id: 2\r\n\r\n'" +
    f" | nc localhost {ts.Variables.port} >> client.log" + " ; echo '======' >> client.log")
tr.Processes.Default.ReturnCode = 0

tr = Test.AddTestRun("bad-http-version")
tr.Processes.Default.Command = (
    f"printf 'GET {random_path} HTTP/1.2\r\n" + "Host: example.com\r\n" + "Connection: close\r\n" + "X-Req-Id: 3\r\n\r\n'" +
    f" | nc localhost {ts.Variables.port} >> client.log" + " ; echo '======' >> client.log")
tr.Processes.Default.ReturnCode = 0

tr = Test.AddTestRun("bad-request-maybe-http-version")
tr.Processes.Default.Command = (
    f"printf 'GET {random_path} HTTP/0.9\r\n" + "Host: example.com\r\n" + "Connection: close\r\n" + "X-Req-Id: 4\r\n\r\n'" +
    f" | nc localhost {ts.Variables.port} >> client.log" + " ; echo '======' >> client.log")
tr.Processes.Default.ReturnCode = 0

tr = Test.AddTestRun("same-as-last-case?")
tr.Processes.Default.Command = (
    fr"printf 'GET {random_path} HTTP/0.9\r\n" + "Host: example.com\r\n" + "Connection: close\r\n" + "X-Req-Id: 5\r\n\r\n'" +
    fr"| nc localhost {ts.Variables.port} >> client.log" + " ; echo '======' >> client.log")
tr.Processes.Default.ReturnCode = 0

tr = Test.AddTestRun("allowed-random-method")
tr.Processes.Default.StartBefore(method_server)
tr.Processes.Default.Command = (
    fr"printf '{random_method} /add-method HTTP/1.1\r\n" + "Host: example.com\r\n" + "Connection: close\r\n" +
    "X-Req-Id: 6\r\n\r\n'" + f" | nc localhost {ts.Variables.port} >> client.log" + " ; echo '======' >> client.log")
tr.Processes.Default.ReturnCode = 0

tr = Test.AddTestRun("valid-ipv6")
tr.Processes.Default.Command = (
    f"printf 'GET /example/2 HTTP/1.1\r\n" + "Host: example.com\r\n" + "Connection: close\r\n" + "X-Req-Id: 7\r\n\r\n'" +
    f" | nc ::1 {ts.Variables.portv6} >> client.log" + " ; echo '======' >> client.log")
tr.Processes.Default.ReturnCode = 0

tr = Test.AddTestRun("unallowed-method-v6")
tr.Processes.Default.Command = (
    f"printf '{random_method} /example/1 HTTP/1.1\r\n" + "Host: example.com\r\n" + "Connection: close\r\n\r\n'" +
    f" | nc ::1 {ts.Variables.portv6} >> client.log" + " ; echo '======' >> client.log")
tr.Processes.Default.ReturnCode = 0

tr = Test.AddTestRun()
tr.Processes.Default.Command = "grep -e '^===' -e '^HTTP/' -e 'X-Resp-Id:' -e '<HTML>' client.log"
tr.Processes.Default.Streams.stdout = 'client.gold'
tr.Processes.Default.ReturnCode = 0

server.Streams.All += Testers.ContainsExpression("Serving GET /sdfsdf/0... Finished", "Served 1")
server.Streams.All += Testers.ContainsExpression("Serving GET /sdfsdf/1... Finished", "Served 2")
server.Streams.All += Testers.ContainsExpression("Serving GET /example/2... Finished", "Served 3")

outserver = Test.Disk.File("outserver")
outserver.Content = "server.gold"



'''
Verify that ATS can perform a reverse lookup when necessary
'''
#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from ports import get_port
import os

Test.Summary = '''
Verify ATS can perform a reverse lookup when necessary
'''

#
# This test verifies a fix for a regression that incorrectly marked a hostdb record as failed doing a reverse DNS
# lookup from the hosts file.  If a parent.config exists, then ATS will perform a reverse lookup on remaps that
# have an ip address in the map-to url. While this doesn't fail the initil request, subsequent requests that resolve
# to the same host record will see the failed lookup attempt and fail the request with a host not found.
#
# This test verfiies the correct behavior in case future code chages introduce a similar regression
#


class DNSReverseLookupTest:
    replay_file = "replay/reverse_lookup.replay.yaml"

    def __init__(self):
        self._server = Test.MakeVerifierServerProcess("server", DNSReverseLookupTest.replay_file)
        self._configure_trafficserver()

    def _configure_trafficserver(self):
        self._ts = Test.MakeATSProcess("ts", enable_cache=False)

        # This first rule would trigger the bug
        self._ts.Disk.remap_config.AddLine(f"map /test1 http://127.0.0.1:{self._server.Variables.http_port}/",)
        # This first rule would fail in the presense of the bug, but this test verifies correct behavior
        self._ts.Disk.remap_config.AddLine(f"map /test2 http://localhost:{self._server.Variables.http_port}/",)
        self._ts.Disk.parent_config.AddLine(
            f'dest_domain=. parent=parent_host:{self._ts.Variables.port} round_robin=consistent_hash go_direct=false')
        self._ts.Disk.parent_config.AddLine(
            # this doesn't need to match, just exist so ats will do the reverse lookup
            f'dest_host=other_host scheme=http parent="parent_host:8080"')

        self._ts.Disk.records_config.update(
            {
                'proxy.config.diags.debug.enabled': 1,
                'proxy.config.diags.debug.tags': 'hostdb|dns|http',
                'proxy.config.http.connect_attempts_max_retries': 0,
                'proxy.config.http.connect_attempts_rr_retries': 0,
                'proxy.config.hostdb.fail.timeout': 10,
                'proxy.config.dns.resolv_conf': 'NULL',
                'proxy.config.hostdb.ttl_mode': 1,
                'proxy.config.hostdb.timeout': 2,
                'proxy.config.hostdb.lookup_timeout': 2,
                'proxy.config.http.transaction_no_activity_timeout_in': 2,
                'proxy.config.http.connect_attempts_timeout': 2,
                'proxy.config.hostdb.host_file.interval': 1,
                'proxy.config.hostdb.host_file.path': os.path.join(Test.TestDirectory, "hosts_file"),
            })

    def test_rev_dns(self):
        tr = Test.AddTestRun()

        tr.Processes.Default.StartBefore(self._server)
        tr.Processes.Default.StartBefore(self._ts)

        tr.AddVerifierClientProcess("client", DNSReverseLookupTest.replay_file, http_ports=[self._ts.Variables.port])

    def run(self):
        self.test_rev_dns()


DNSReverseLookupTest().run()



#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import os

Test.Summary = '''
Execute plugin with cppapi tests.
'''

ts = Test.MakeATSProcess("ts")

Test.PrepareTestPlugin(os.path.join(Test.Variables.AtsTestPluginsDir, 'test_cppapi.so'), ts)

tr = Test.AddTestRun()
tr.Processes.Default.StartBefore(Test.Processes.ts)
tr.Processes.Default.Command = "echo run test_cppapi plugin"
tr.Processes.Default.ReturnCode = 0
tr.Processes.StillRunningAfter = ts



"""
"""
#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from typing import Union

Test.Summary = 'Verify ATS can function as a forward proxy'
Test.ContinueOnFail = True


class ForwardProxyTest:
    _scheme_proto_mismatch_policy: Union[int, None]
    _ts_counter: int = 0
    _server_counter: int = 0

    def __init__(self, verify_scheme_matches_protocol: Union[int, None]):
        """Construct a ForwardProxyTest object.

        :param verify_scheme_matches_protocol: The value with which to
        configure Traffic Server's
        proxy.config.ssl.client.scheme_proto_mismatch_policy. A value of None
        means that no value will be explicitly set in the records.yaml.
        :type verify_scheme_matches_protocol: int or None
        """
        self._scheme_proto_mismatch_policy = verify_scheme_matches_protocol
        self.setupOriginServer()
        self.setupTS()

    def setupOriginServer(self):
        """Configure the Proxy Verifier server."""
        proc_name = f"server{ForwardProxyTest._server_counter}"
        self.server = Test.MakeVerifierServerProcess(proc_name, "forward_proxy.replay.yaml")
        ForwardProxyTest._server_counter += 1
        if self._scheme_proto_mismatch_policy in (2, None):
            self.server.Streams.All = Testers.ExcludesExpression(
                'Received an HTTP/1 request with key 1', 'Verify that the server did not receive the request.')
        else:
            self.server.Streams.All = Testers.ContainsExpression(
                'Received an HTTP/1 request with key 1', 'Verify that the server received the request.')

    def setupTS(self):
        """Configure the Traffic Server process."""
        proc_name = f"ts{ForwardProxyTest._ts_counter}"
        self.ts = Test.MakeATSProcess(proc_name, enable_tls=True, enable_cache=False)
        ForwardProxyTest._ts_counter += 1
        self.ts.addDefaultSSLFiles()
        self.ts.Disk.ssl_multicert_config.AddLine("dest_ip=* ssl_cert_name=server.pem ssl_key_name=server.key")
        self.ts.Disk.remap_config.AddLine(f"map / http://127.0.0.1:{self.server.Variables.http_port}/")

        self.ts.Disk.records_config.update(
            {
                'proxy.config.ssl.server.cert.path': self.ts.Variables.SSLDir,
                'proxy.config.ssl.server.private_key.path': self.ts.Variables.SSLDir,
                'proxy.config.ssl.client.verify.server.policy': 'PERMISSIVE',
                'proxy.config.diags.debug.enabled': 1,
                'proxy.config.diags.debug.tags': "http",
            })

        if self._scheme_proto_mismatch_policy is not None:
            self.ts.Disk.records_config.update(
                {
                    'proxy.config.ssl.client.scheme_proto_mismatch_policy': self._scheme_proto_mismatch_policy,
                })

    def addProxyHttpsToHttpCase(self):
        """Test ATS as an HTTPS forward proxy behind an HTTP server."""
        tr = Test.AddTestRun()
        tr.Processes.Default.StartBefore(self.server)
        tr.Processes.Default.StartBefore(self.ts)
        tr.Processes.Default.Command = (
            f'curl --proxy-insecure -v -H "uuid: 1" '
            f'--proxy "https://127.0.0.1:{self.ts.Variables.ssl_port}/" '
            f'http://example.com/')
        tr.Processes.Default.ReturnCode = 0
        tr.StillRunningAfter = self.server
        tr.StillRunningAfter = self.ts

        if self._scheme_proto_mismatch_policy in (2, None):
            tr.Processes.Default.Streams.All = Testers.ContainsExpression(
                '< HTTP/1.1 400 Invalid HTTP Request', 'Verify that the request was rejected.')
        else:
            tr.Processes.Default.Streams.All = Testers.ContainsExpression(
                '< HTTP/1.1 200 OK', 'Verify that curl received a 200 OK response.')

    def run(self):
        """Configure the TestRun instances for this set of tests."""
        self.addProxyHttpsToHttpCase()


ForwardProxyTest(verify_scheme_matches_protocol=None).run()
ForwardProxyTest(verify_scheme_matches_protocol=0).run()
ForwardProxyTest(verify_scheme_matches_protocol=1).run()
ForwardProxyTest(verify_scheme_matches_protocol=2).run()



'''
Verify IP allow ip_category behavior.
'''
#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import os
import re

from typing import List

Test.Summary = '''
Verify IP allow ip_category behavior.
'''


class CategoryFile:
    """Encapsulate the various ip_category.yaml contents."""

    contents: List['CategoryFile'] = []
    parent_directory: str = Test.RunDirectory
    _index: int = 0

    def __init__(self, content: str):
        """Initialize the object.

        :param content: The content of the ip_category.yaml file.
        """
        self._content = content
        self._index = len(CategoryFile.contents)
        self._filename = os.path.join(CategoryFile.parent_directory, f'categories{self._index}.yaml')
        CategoryFile.contents.append(self)

    def _write(self):
        with open(self._filename, 'w') as f:
            f.write(self._content)

    def get_path(self):
        return self._filename

    @classmethod
    def write_all(cls):
        for content in cls.contents:
            content._write()


localhost_is_internal_and_external = CategoryFile(
    '''
ip_categories:
  - name: ACME_INTERNAL
    ip_addrs: 127.0.0.1
  - name: ACME_EXTERNAL
    ip_addrs: 127.0.0.1
  - name: ACME_ALL
    ip_addrs: 127.0.0.1
  - name: ALL
    ip_addrs: 127.0.0.1
''')

localhost_is_external = CategoryFile(
    '''
ip_categories:
  - name: ACME_INTERNAL
    ip_addrs: 1.2.3.4
  - name: ACME_REMAP_EXTERNAL
    ip_addrs: 127.0.0.1
  - name: ACME_EXTERNAL
    ip_addrs: 127.0.0.1
  - name: ACME_ALL
    ip_addrs:
      - 1.2.3.4
      - 127.0.0.1
  - name: ALL
    ip_addrs:
      - 1.2.3.4
      - 127.0.0.1
''')

localhost_is_neither = CategoryFile(
    '''
ip_categories:
  - name: ACME_INTERNAL
    ip_addrs: 1.2.3.4
  - name: ACME_EXTERNAL
    ip_addrs: 1.2.3.4
  - name: ACME_ALL
    ip_addrs: 1.2.3.4
  - name: ALL
    ip_addrs:
      - 1.2.3.4
      - 127.0.0.1
''')

# Keep this below the above content instantiations.
CategoryFile.write_all()


class Test_ip_category:
    """Configure a test to verify ip_category behavior."""

    _client_counter: int = 0
    _ts_is_started: bool = False
    _reload_server_is_started: bool = False
    _server_is_started: bool = False
    _reload_counter: int = 0

    _ts: 'TestProcess' = None
    _server: 'TestProcess' = None
    _reload_server: 'TestProcess' = None

    _categories_filename: str = f'{Test.RunDirectory}/categories.yaml'
    _category_files_are_written: bool = False

    _server_replay = 'replays/https_categories_server.replay.yaml'

    def __init__(
            self, name: str, replay_file: str, ip_allow_config: str, ip_category_config: 'CategoryFile', acl_configuration: str,
            expected_responses: List[int]):
        """Initialize the test.

        :param name: The name of the test.
        :param replay_file: The replay file to be used.
        :param ip_allow_config: The ip_allow configuration to be used.
        :param ip_category_config: The ip_category.yaml configuration to be used.
        :param acl_configuration: The ACL configuration to be used.
        :param expect_responses: The in-order expected responses from the proxy.
        """
        self._replay_file = replay_file
        self._ip_allow_config = ip_allow_config
        self._acl_configuration = acl_configuration
        self._expected_responses = expected_responses

        self._update_categories_file(ip_category_config)
        self._update_remap_with_acl()

        self._configure_server()
        self._configure_traffic_server()

        tr = Test.AddTestRun(name)
        self._configure_client(tr)

    def _update_remap_with_acl(self) -> None:
        """Update the remap.config file with the ACL configuration."""
        if Test_ip_category._ts:
            if self._acl_configuration:
                tr = Test.AddTestRun(f"remap.config file update with acl: {self._acl_configuration}")
                p = tr.Processes.Default
                destination = os.path.join(Test_ip_category._ts.Variables.CONFIGDIR, 'remap.config')
                common = f'map / http://127.0.0.1:{Test_ip_category._server.Variables.http_port} '
                p.Command = f'echo {common} {self._acl_configuration} > {destination}; cat {destination}'
                p.ReturnCode = 0
            else:
                tr = Test.AddTestRun(f"remap.config file update with no acl")
                p = tr.Processes.Default
                destination = os.path.join(Test_ip_category._ts.Variables.CONFIGDIR, 'remap.config')
                common = f'map / http://127.0.0.1:{Test_ip_category._server.Variables.http_port} '
                p.Command = f'echo {common} > {destination}; cat {destination}'
                p.ReturnCode = 0

    def _update_categories_file(self, category_content: 'CategoryFile') -> None:
        """Update the categories file.

        :param category_content: The content of the categories file.
        """
        tr = Test.AddTestRun(f"Categories file update: {category_content.get_path()}")
        p = tr.Processes.Default
        destination = Test_ip_category._categories_filename
        p.Command = f'cp {category_content.get_path()} {destination}; cat {destination}; ls -ltr {destination}'
        p.ReturnCode = 0

    def _configure_server(self) -> None:
        """Configure the server."""
        if Test_ip_category._server:
            # All test runs share a single server instance.
            return
        server = Test.MakeVerifierServerProcess(f"server", self._server_replay)
        Test_ip_category._server = server

    def _configure_traffic_server(self) -> None:
        """Configure Traffic Server."""

        if Test_ip_category._ts:
            # All test runs share a single Traffic Server instance.

            # Reload the ip_allow.yaml file.
            ts = Test_ip_category._ts
            tr = Test.AddTestRun(f"Reload the configuration file.")
            Test_ip_category._reload_counter += 1
            p = tr.Processes.Process(f"reload-{Test_ip_category._reload_counter}")
            # The sleep is added to give time for the reload to happen.
            p.Command = 'traffic_ctl config reload; sleep 30'
            p.Env = ts.Env
            # Killing the sleep can result in a -2 return code.
            p.ReturnCode = Any(0, -2)
            p.Ready = When.FileContains(
                ts.Disk.diags_log.Name, "ip_allow.yaml finished loading", 1 + Test_ip_category._reload_counter)
            p.Timeout = 20
            tr.StillRunningAfter = ts
            tr.Processes.Default.StartBefore(p)
            tr.Processes.Default.Command = 'echo "waiting upon traffic server to reload"'
            tr.TimeOut = 20

            return
        ts = Test.MakeATSProcess("ts", enable_cache=False, enable_tls=True)
        Test_ip_category._ts = ts

        ts.addDefaultSSLFiles()
        ts.Disk.ssl_multicert_config.AddLine('dest_ip=* ssl_cert_name=server.pem ssl_key_name=server.key')
        ts.Disk.records_config.update(
            {
                'proxy.config.diags.debug.enabled': 1,
                'proxy.config.diags.debug.tags': 'http|ip_allow',
                'proxy.config.cache.ip_categories.filename': Test_ip_category._categories_filename,
                'proxy.config.http.push_method_enabled': 1,
                'proxy.config.ssl.server.cert.path': ts.Variables.SSLDir,
                'proxy.config.quic.no_activity_timeout_in': 0,
                'proxy.config.ssl.server.private_key.path': ts.Variables.SSLDir,
                'proxy.config.ssl.client.verify.server.policy': 'PERMISSIVE',
                'proxy.config.http.connect_ports': Test_ip_category._server.Variables.http_port,
                'proxy.config.url_remap.acl_behavior_policy': 1,  # TODO: adjust expected_responses with the default config
            })

        ts.Disk.remap_config.AddLine(
            f'map / http://127.0.0.1:{Test_ip_category._server.Variables.http_port} {self._acl_configuration}')
        ts.Disk.ip_allow_yaml.AddLines(self._ip_allow_config.split("\n"))

    def _configure_client(self, tr: 'TestRun') -> None:
        """Run the test.

        :param tr: The TestRun object to associate the client process with.
        """

        if not Test_ip_category._server_is_started:
            tr.Processes.Default.StartBefore(Test_ip_category._server)
            Test_ip_category._server_is_started = True
        if not Test_ip_category._ts_is_started:
            tr.Processes.Default.StartBefore(Test_ip_category._ts)
            Test_ip_category._ts_is_started = True

        p = tr.AddVerifierClientProcess(
            f'client-{Test_ip_category._client_counter}', self._replay_file, https_ports=[Test_ip_category._ts.Variables.ssl_port])
        Test_ip_category._client_counter += 1

        if self._expected_responses:
            codes = [str(code) for code in self._expected_responses]
            p.Streams.stdout += Testers.ContainsExpression(
                '.*'.join(codes), "Verifying the expected order of responses", reflags=re.DOTALL | re.MULTILINE)
        else:
            # If there are no expected responses, expect the Warning about the rejected ip.
            self._ts.Disk.diags_log.Content += Testers.ContainsExpression(
                "client '127.0.0.1' prohibited by ip-allow policy", "Verify the client rejection warning message.")

            # Also, the client will complain about the broken connections.
            p.ReturnCode = 1


IP_ALLOW_CONTENT = f'''
ip_allow:
  - apply: in
    ip_categories: ACME_INTERNAL
    action: set_allow
    methods:
      - GET
      - HEAD
      - POST
      - PUSH
  - apply: in
    ip_categories: ACME_EXTERNAL
    action: set_allow
    methods:
      - GET
      - HEAD
  - apply: in
    ip_categories: ACME_ALL
    action: set_allow
    methods:
      - HEAD
  - apply: in
    ip_categories: ALL
    action: set_deny
'''

test_ip_allow_optional_methods = Test_ip_category(
    "IP Category: INTERNAL",
    replay_file='replays/https_categories_internal.replay.yaml',
    ip_allow_config=IP_ALLOW_CONTENT,
    ip_category_config=localhost_is_internal_and_external,
    acl_configuration='',
    expected_responses=[200, 200, 400, 403])

test_ip_allow_optional_methods = Test_ip_category(
    "IP Category: EXTERNAL",
    replay_file='replays/https_categories_external.replay.yaml',
    ip_allow_config=IP_ALLOW_CONTENT,
    ip_category_config=localhost_is_external,
    acl_configuration='',
    expected_responses=[200, 403, 403])

# Because all requests are outright rejected for 127.0.0.1, ATS will
# reject all incoming transactions and not even give a 403 response.
test_ip_allow_optional_methods = Test_ip_category(
    "IP Category: ALL",
    replay_file='replays/https_categories_all.replay.yaml',
    ip_allow_config=IP_ALLOW_CONTENT,
    ip_category_config=localhost_is_neither,
    acl_configuration='',
    expected_responses=None)

# Deny GET via remap.config ACL.
test_ip_allow_optional_methods = Test_ip_category(
    "IP Category: EXTERNAL",
    replay_file='replays/https_categories_external_remap.replay.yaml',
    ip_allow_config=IP_ALLOW_CONTENT,
    ip_category_config=localhost_is_external,
    acl_configuration='@action=set_deny @src_ip_category=ACME_REMAP_EXTERNAL @method=GET',
    expected_responses=[403, 200, 200])



'''
'''
#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import sys

Test.Summary = 'Test that Trafficserver starts with different thread configurations.'
Test.ContinueOnFail = True

ts = Test.MakeATSProcess('ts-1_exec-0_accept-1_task-1_aio')
ts.Disk.records_config.update(
    {
        'proxy.config.exec_thread.autoconfig.enabled': 0,
        'proxy.config.exec_thread.autoconfig.scale': 1.5,
        'proxy.config.exec_thread.limit': 1,
        'proxy.config.accept_threads': 0,
        'proxy.config.task_threads': 1,
        'proxy.config.cache.threads_per_disk': 1,
        'proxy.config.diags.debug.enabled': 1,
        'proxy.config.diags.debug.tags': 'iocore_thread_start|iocore_net_accept_start'
    })
ts.Setup.CopyAs('check_threads.py', Test.RunDirectory)

tr = Test.AddTestRun()
TS_ROOT = ts.Env['TS_ROOT']
tr.Processes.Default.Command = f'{sys.executable} check_threads.py -p {TS_ROOT} -e 1 -a 0 -t 1 -c 1'
tr.Processes.Default.ReturnCode = 0
tr.Processes.Default.StartBefore(ts)

ts = Test.MakeATSProcess('ts-1_exec-1_accept-2_task-8_aio')
ts.Disk.records_config.update(
    {
        'proxy.config.exec_thread.autoconfig.enabled': 0,
        'proxy.config.exec_thread.autoconfig.scale': 1.5,
        'proxy.config.exec_thread.limit': 1,
        'proxy.config.accept_threads': 1,
        'proxy.config.task_threads': 2,
        'proxy.config.cache.threads_per_disk': 8,
        'proxy.config.diags.debug.enabled': 1,
        'proxy.config.diags.debug.tags': 'iocore_thread_start|iocore_net_accept_start'
    })

tr = Test.AddTestRun()
TS_ROOT = ts.Env['TS_ROOT']
tr.Processes.Default.Command = f'{sys.executable} check_threads.py -p {TS_ROOT} -e 1 -a 1 -t 2 -c 8'
tr.Processes.Default.ReturnCode = 0
tr.Processes.Default.StartBefore(ts)

ts = Test.MakeATSProcess('ts-1_exec-10_accept-10_task-32_aio')
ts.Disk.records_config.update(
    {
        'proxy.config.exec_thread.autoconfig.enabled': 0,
        'proxy.config.exec_thread.autoconfig.scale': 1.5,
        'proxy.config.exec_thread.limit': 1,
        'proxy.config.accept_threads': 10,
        'proxy.config.task_threads': 10,
        'proxy.config.cache.threads_per_disk': 32,
        'proxy.config.diags.debug.enabled': 1,
        'proxy.config.diags.debug.tags': 'iocore_thread_start|iocore_net_accept_start'
    })

tr = Test.AddTestRun()
TS_ROOT = ts.Env['TS_ROOT']
tr.Processes.Default.Command = f'{sys.executable} check_threads.py -p {TS_ROOT} -e 1 -a 10 -t 10 -c 32'
tr.Processes.Default.ReturnCode = 0
tr.Processes.Default.StartBefore(ts)

ts = Test.MakeATSProcess('ts-2_exec-0_accept-1_task-1_aio')
ts.Disk.records_config.update(
    {
        'proxy.config.exec_thread.autoconfig.enabled': 0,
        'proxy.config.exec_thread.autoconfig.scale': 1.5,
        'proxy.config.exec_thread.limit': 2,
        'proxy.config.accept_threads': 0,
        'proxy.config.task_threads': 1,
        'proxy.config.cache.threads_per_disk': 1,
        'proxy.config.diags.debug.enabled': 1,
        'proxy.config.diags.debug.tags': 'iocore_thread_start|iocore_net_accept_start'
    })

tr = Test.AddTestRun()
TS_ROOT = ts.Env['TS_ROOT']
tr.Processes.Default.Command = f'{sys.executable} check_threads.py -p {TS_ROOT} -e 2 -a 0 -t 1 -c 1'
tr.Processes.Default.ReturnCode = 0
tr.Processes.Default.StartBefore(ts)

ts = Test.MakeATSProcess('ts-2_exec-1_accept-2_task-8_aio')
ts.Disk.records_config.update(
    {
        'proxy.config.exec_thread.autoconfig.enabled': 0,
        'proxy.config.exec_thread.autoconfig.scale': 1.5,
        'proxy.config.exec_thread.limit': 2,
        'proxy.config.accept_threads': 1,
        'proxy.config.task_threads': 2,
        'proxy.config.cache.threads_per_disk': 8,
        'proxy.config.diags.debug.enabled': 1,
        'proxy.config.diags.debug.tags': 'iocore_thread_start|iocore_net_accept_start'
    })

tr = Test.AddTestRun()
TS_ROOT = ts.Env['TS_ROOT']
tr.Processes.Default.Command = f'{sys.executable} check_threads.py -p {TS_ROOT} -e 2 -a 1 -t 2 -c 8'
tr.Processes.Default.ReturnCode = 0
tr.Processes.Default.StartBefore(ts)

ts = Test.MakeATSProcess('ts-2_exec-10_accept-10_task-32_aio')
ts.Disk.records_config.update(
    {
        'proxy.config.exec_thread.autoconfig.enabled': 0,
        'proxy.config.exec_thread.autoconfig.scale': 1.5,
        'proxy.config.exec_thread.limit': 2,
        'proxy.config.accept_threads': 10,
        'proxy.config.task_threads': 10,
        'proxy.config.cache.threads_per_disk': 32,
        'proxy.config.diags.debug.enabled': 1,
        'proxy.config.diags.debug.tags': 'iocore_thread_start|iocore_net_accept_start'
    })

tr = Test.AddTestRun()
TS_ROOT = ts.Env['TS_ROOT']
tr.Processes.Default.Command = f'{sys.executable} check_threads.py -p {TS_ROOT} -e 2 -a 10 -t 10 -c 32'
tr.Processes.Default.ReturnCode = 0
tr.Processes.Default.StartBefore(ts)

ts = Test.MakeATSProcess('ts-32_exec-0_accept-1_task-1_aio')
ts.Disk.records_config.update(
    {
        'proxy.config.exec_thread.autoconfig.enabled': 0,
        'proxy.config.exec_thread.autoconfig.scale': 1.5,
        'proxy.config.exec_thread.limit': 32,
        'proxy.config.accept_threads': 0,
        'proxy.config.task_threads': 1,
        'proxy.config.cache.threads_per_disk': 1,
        'proxy.config.diags.debug.enabled': 1,
        'proxy.config.diags.debug.tags': 'iocore_thread_start|iocore_net_accept_start'
    })

tr = Test.AddTestRun()
TS_ROOT = ts.Env['TS_ROOT']
tr.Processes.Default.Command = f'{sys.executable} check_threads.py -p {TS_ROOT} -e 32 -a 0 -t 1 -c 1'
tr.Processes.Default.ReturnCode = 0
tr.Processes.Default.StartBefore(ts)

ts = Test.MakeATSProcess('ts-32_exec-1_accept-2_task-8_aio')
ts.Disk.records_config.update(
    {
        'proxy.config.exec_thread.autoconfig.enabled': 0,
        'proxy.config.exec_thread.autoconfig.scale': 1.5,
        'proxy.config.exec_thread.limit': 32,
        'proxy.config.accept_threads': 1,
        'proxy.config.task_threads': 2,
        'proxy.config.cache.threads_per_disk': 8,
        'proxy.config.diags.debug.enabled': 1,
        'proxy.config.diags.debug.tags': 'iocore_thread_start|iocore_net_accept_start'
    })

tr = Test.AddTestRun()
TS_ROOT = ts.Env['TS_ROOT']
tr.Processes.Default.Command = f'{sys.executable} check_threads.py -p {TS_ROOT} -e 32 -a 1 -t 2 -c 8'
tr.Processes.Default.ReturnCode = 0
tr.Processes.Default.StartBefore(ts)

ts = Test.MakeATSProcess('ts-32_exec-10_accept-10_task-32_aio')
ts.Disk.records_config.update(
    {
        'proxy.config.exec_thread.autoconfig.enabled': 0,
        'proxy.config.exec_thread.autoconfig.scale': 1.5,
        'proxy.config.exec_thread.limit': 32,
        'proxy.config.accept_threads': 10,
        'proxy.config.task_threads': 10,
        'proxy.config.cache.threads_per_disk': 32,
        'proxy.config.diags.debug.enabled': 1,
        'proxy.config.diags.debug.tags': 'iocore_thread_start|iocore_net_accept_start'
    })

tr = Test.AddTestRun()
TS_ROOT = ts.Env['TS_ROOT']
tr.Processes.Default.Command = f'{sys.executable} check_threads.py -p {TS_ROOT} -e 32 -a 10 -t 10 -c 32'
tr.Processes.Default.ReturnCode = 0
tr.Processes.Default.StartBefore(ts)

ts = Test.MakeATSProcess('ts-100_exec-0_accept-1_task-1_aio')
ts.Disk.records_config.update(
    {
        'proxy.config.exec_thread.autoconfig.enabled': 0,
        'proxy.config.exec_thread.autoconfig.scale': 1.5,
        'proxy.config.exec_thread.limit': 100,
        'proxy.config.accept_threads': 0,
        'proxy.config.task_threads': 1,
        'proxy.config.cache.threads_per_disk': 1,
        'proxy.config.diags.debug.enabled': 1,
        'proxy.config.diags.debug.tags': 'iocore_thread_start|iocore_net_accept_start'
    })

tr = Test.AddTestRun()
TS_ROOT = ts.Env['TS_ROOT']
tr.Processes.Default.Command = f'{sys.executable} check_threads.py -p {TS_ROOT} -e 100 -a 0 -t 1 -c 1'
tr.Processes.Default.ReturnCode = 0
tr.Processes.Default.StartBefore(ts)

ts = Test.MakeATSProcess('ts-100_exec-1_accept-2_task-8_aio')
ts.Disk.records_config.update(
    {
        'proxy.config.exec_thread.autoconfig.enabled': 0,
        'proxy.config.exec_thread.autoconfig.scale': 1.5,
        'proxy.config.exec_thread.limit': 100,
        'proxy.config.accept_threads': 1,
        'proxy.config.task_threads': 2,
        'proxy.config.cache.threads_per_disk': 8,
        'proxy.config.diags.debug.enabled': 1,
        'proxy.config.diags.debug.tags': 'iocore_thread_start|iocore_net_accept_start'
    })

tr = Test.AddTestRun()
TS_ROOT = ts.Env['TS_ROOT']
tr.Processes.Default.Command = f'{sys.executable} check_threads.py -p {TS_ROOT} -e 100 -a 1 -t 2 -c 8'
tr.Processes.Default.ReturnCode = 0
tr.Processes.Default.StartBefore(ts)

ts = Test.MakeATSProcess('ts-100_exec-10_accept-10_task-32_aio')
ts.Disk.records_config.update(
    {
        'proxy.config.exec_thread.autoconfig.enabled': 0,
        'proxy.config.exec_thread.autoconfig.scale': 1.5,
        'proxy.config.exec_thread.limit': 100,
        'proxy.config.accept_threads': 10,
        'proxy.config.task_threads': 10,
        'proxy.config.cache.threads_per_disk': 32,
        'proxy.config.diags.debug.enabled': 1,
        'proxy.config.diags.debug.tags': 'iocore_thread_start|iocore_net_accept_start'
    })

tr = Test.AddTestRun()
TS_ROOT = ts.Env['TS_ROOT']
tr.Processes.Default.Command = f'{sys.executable} check_threads.py -p {TS_ROOT} -e 100 -a 10 -t 10 -c 32'
tr.Processes.Default.ReturnCode = 0
tr.Processes.Default.StartBefore(ts)



'''
'''
#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import os

Test.Summary = '''
Test unsupported values for chunked_encoding
'''

Test.ContinueOnFail = True

# Define default ATS
ts = Test.MakeATSProcess("ts1")
server = Test.MakeOriginServer("server")

testName = ""
request_header = {
    "headers": "POST /case1 HTTP/1.1\r\nHost: www.example.com\r\nuuid:1\r\n\r\n",
    "timestamp": "1469733493.993",
    "body": "stuff"
}
response_header = {
    "headers": "HTTP/1.1 200 OK\r\nServer: uServer\r\nConnection: close\r\nTransfer-Encoding: chunked\r\n\r\n",
    "timestamp": "1469733493.993",
    "body": "more stuff"
}

server.addResponse("sessionlog.json", request_header, response_header)

ts.Disk.records_config.update({'proxy.config.diags.debug.enabled': 0, 'proxy.config.diags.debug.tags': 'http'})

ts.Disk.remap_config.AddLine('map / http://127.0.0.1:{0}'.format(server.Variables.Port))

# HTTP1.1 POST: www.example.com/case1 with gzip transfer-encoding
tr = Test.AddTestRun()
tr.TimeOut = 5
tr.Processes.Default.Command = 'curl -H "host: example.com" -H "transfer-encoding: gzip" -d "stuff" http://127.0.0.1:{0}/case1  --verbose'.format(
    ts.Variables.port)
tr.Processes.Default.ReturnCode = 0
tr.Processes.Default.StartBefore(server)
tr.Processes.Default.StartBefore(ts)
tr.Processes.Default.Streams.All = Testers.ContainsExpression("501 Field not implemented", "Should fail")
tr.Processes.Default.Streams.All = Testers.ExcludesExpression("200 OK", "Should not succeed")
tr.StillRunningAfter = server
tr.StillRunningAfter = ts

# HTTP1.1 POST: www.example.com/case1 with gzip and chunked transfer-encoding
tr = Test.AddTestRun()
tr.TimeOut = 5
tr.Processes.Default.Command = 'curl -H "host: example.com" -H "transfer-encoding: gzip" -H "transfer-encoding: chunked" -d "stuff" http://127.0.0.1:{0}/case1  --verbose'.format(
    ts.Variables.port)
tr.Processes.Default.ReturnCode = 0
tr.Processes.Default.Streams.All = Testers.ContainsExpression("501 Field not implemented", "Should fail")
tr.Processes.Default.Streams.All = Testers.ExcludesExpression("200 OK", "Should not succeed")
tr.StillRunningAfter = server
tr.StillRunningAfter = ts


class HTTP10Test:
    chunkedReplayFile = "replays/chunked_in_http_1_0.replay.yaml"

    def __init__(self):
        self.setupOriginServer()
        self.setupTS()

    def setupOriginServer(self):
        self.server = Test.MakeVerifierServerProcess("verifier-server1", self.chunkedReplayFile)

    def setupTS(self):
        self.ts = Test.MakeATSProcess("ts2", enable_tls=True, enable_cache=False)
        self.ts.addDefaultSSLFiles()
        self.ts.Disk.records_config.update(
            {
                "proxy.config.diags.debug.enabled": 1,
                "proxy.config.diags.debug.tags": "http",
                "proxy.config.ssl.server.cert.path": f'{self.ts.Variables.SSLDir}',
                "proxy.config.ssl.server.private_key.path": f'{self.ts.Variables.SSLDir}',
                "proxy.config.ssl.client.verify.server.policy": 'PERMISSIVE',
            })
        self.ts.Disk.ssl_multicert_config.AddLine('dest_ip=* ssl_cert_name=server.pem ssl_key_name=server.key')
        self.ts.Disk.remap_config.AddLine(f"map / http://127.0.0.1:{self.server.Variables.http_port}/",)

    def runChunkedTraffic(self):
        tr = Test.AddTestRun()
        tr.AddVerifierClientProcess(
            "client1",
            self.chunkedReplayFile,
            http_ports=[self.ts.Variables.port],
            https_ports=[self.ts.Variables.ssl_port],
            other_args='--thread-limit 1')
        tr.Processes.Default.StartBefore(self.server)
        tr.Processes.Default.StartBefore(self.ts)
        tr.StillRunningAfter = self.server
        tr.StillRunningAfter = self.ts

    def run(self):
        self.runChunkedTraffic()


HTTP10Test().run()


class MalformedChunkHeaderTest:
    chunkedReplayFile = "replays/malformed_chunked_header.replay.yaml"

    def __init__(self):
        self.setupOriginServer()
        self.setupTS()

    def setupOriginServer(self):
        self.server = Test.MakeVerifierServerProcess("verifier-server2", self.chunkedReplayFile)

        # The server's responses will fail the first three transactions
        # because ATS will close the connection due to the malformed
        # chunk headers.
        self.server.Streams.stdout += Testers.ContainsExpression(
            "Unexpected chunked content for key 1: too small", "Verify that writing the first response failed.")
        self.server.Streams.stdout += Testers.ExcludesExpression(
            "chunked body of 3 bytes for key 2 with chunk stream", "Verify that writing the second response failed.")
        self.server.Streams.stdout += Testers.ContainsExpression(
            "Unexpected chunked content for key 3: too small", "Verify that writing the third response failed.")

        # ATS should close the connection before any body gets through. "abcwxyz"
        # is the body sent by the client for each of these chunked cases.
        self.server.Streams.stdout += Testers.ExcludesExpression("abcwxyz", "Verify that the body never got through.")

    def setupTS(self):
        self.ts = Test.MakeATSProcess("ts3", enable_tls=True, enable_cache=False)
        self.ts.addDefaultSSLFiles()
        self.ts.Disk.records_config.update(
            {
                "proxy.config.diags.debug.enabled": 1,
                "proxy.config.diags.debug.tags": "http",
                "proxy.config.ssl.server.cert.path": f'{self.ts.Variables.SSLDir}',
                "proxy.config.ssl.server.private_key.path": f'{self.ts.Variables.SSLDir}',
                "proxy.config.ssl.client.verify.server.policy": 'PERMISSIVE',
            })
        self.ts.Disk.ssl_multicert_config.AddLine('dest_ip=* ssl_cert_name=server.pem ssl_key_name=server.key')
        self.ts.Disk.remap_config.AddLine(f"map / http://127.0.0.1:{self.server.Variables.http_port}/",)
        self.ts.Disk.traffic_out.Content += Testers.ContainsExpression(
            "user agent post chunk decoding error", "Verify that ATS detected a problem parsing a chunk.")

    def runChunkedTraffic(self):
        tr = Test.AddTestRun()
        tr.AddVerifierClientProcess(
            "client2",
            self.chunkedReplayFile,
            http_ports=[self.ts.Variables.port],
            https_ports=[self.ts.Variables.ssl_port],
            other_args='--thread-limit 1')
        tr.Processes.Default.StartBefore(self.server)
        tr.Processes.Default.StartBefore(self.ts)
        tr.StillRunningAfter = self.server
        tr.StillRunningAfter = self.ts

        # The aborted connections will result in errors and a non-zero return
        # code from the verifier client.
        tr.Processes.Default.ReturnCode = 1
        tr.Processes.Default.Streams.stdout += Testers.ContainsExpression(
            r"(Unexpected chunked content for key 4: too small|Failed HTTP/1 transaction with key: 4)",
            "Verify that ATS closed the forth transaction.")
        tr.Processes.Default.Streams.stdout += Testers.ContainsExpression(
            r"(Unexpected chunked content for key 5: too small|Failed HTTP/1 transaction with key: 5)",
            "Verify that ATS closed the fifth transaction.")
        tr.Processes.Default.Streams.stdout += Testers.ContainsExpression(
            r"(Unexpected chunked content for key 6: too small|Failed HTTP/1 transaction with key: 6)",
            "Verify that ATS closed the sixth transaction.")

        # ATS should close the connection before any body gets through. "def"
        # is the body sent by the server for each of these chunked cases.
        tr.Processes.Default.Streams.stdout += Testers.ExcludesExpression("def", "Verify that the body never got through.")

    def run(self):
        self.runChunkedTraffic()


MalformedChunkHeaderTest().run()



#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
import os

Test.Summary = '''
Test remap reloading
'''
Test.testName = 'remap_reload'

replay_file_1 = "reload_1.replay.yaml"
replay_file_2 = "reload_2.replay.yaml"
replay_file_3 = "reload_3.replay.yaml"
replay_file_4 = "reload_4.replay.yaml"

tm = Test.MakeATSProcess("ts")
tm.Disk.diags_log.Content = Testers.ContainsExpression("remap.config failed to load", "Remap should fail to load")
remap_cfg_path = os.path.join(tm.Variables.CONFIGDIR, 'remap.config')

pv = Test.MakeVerifierServerProcess("pv", "reload_server.replay.yaml")
pv_port = pv.Variables.http_port
tm.Disk.remap_config.AddLines(
    [
        f"map http://alpha.ex http://alpha.ex:{pv_port}",
        f"map http://bravo.ex http://bravo.ex:{pv_port}",
        f"map http://charlie.ex http://charlie.ex:{pv_port}",
        f"map http://delta.ex http://delta.ex:{pv_port}",
    ])
tm.Disk.records_config.update({'proxy.config.url_remap.min_rules_required': 3})

nameserver = Test.MakeDNServer("dns", default='127.0.0.1')
tm.Disk.records_config.update(
    {
        'proxy.config.dns.nameservers': f"127.0.0.1:{nameserver.Variables.Port}",
        'proxy.config.dns.resolv_conf': 'NULL'
    })

tr = Test.AddTestRun("verify load")
tr.Processes.Default.StartBefore(pv)
tr.Processes.Default.StartBefore(nameserver)
tr.Processes.Default.StartBefore(tm)
tr.AddVerifierClientProcess("client", replay_file_1, http_ports=[tm.Variables.port])

tr = Test.AddTestRun("Change remap.config to have only two lines")
tr.Processes.Default.Env = tm.Env
tr.Processes.Default.Command = 'echo "Change remap.config, two lines"'
tr.Disk.File(remap_cfg_path).WriteOn("")
tr.Disk.File(
    remap_cfg_path, typename="ats:config").AddLines(
        [
            f"map http://alpha.ex http://alpha.ex:{pv_port}",
            f"map http://bravo.ex http://bravo.ex:{pv_port}",
        ])

tr = Test.AddTestRun("remap_config reload, fails")
tr.Processes.Default.Env = tm.Env
tr.Processes.Default.Command = 'sleep 2; traffic_ctl config reload'

tr = Test.AddTestRun("after first reload")
await_config_reload = tr.Processes.Process('config_reload_failed', 'sleep 30')
await_config_reload.Ready = When.FileContains(tm.Disk.diags_log.Name, "configuration is invalid")
tr.AddVerifierClientProcess("client_2", replay_file_2, http_ports=[tm.Variables.port])
tr.Processes.Default.StartBefore(await_config_reload)

tr = Test.AddTestRun("Change remap.config to have more than three lines")
tr.Processes.Default.Env = tm.Env
tr.Processes.Default.Command = 'echo "Change remap.config, more than three lines"'
tr.Disk.File(remap_cfg_path).WriteOn("")
tr.Disk.File(
    remap_cfg_path, typename="ats:config").AddLines(
        [
            f"map http://echo.ex http://echo.ex:{pv_port}",
            f"map http://foxtrot.ex http://foxtrot.ex:{pv_port}",
            f"map http://golf.ex http://golf.ex:{pv_port}",
            f"map http://hotel.ex http://hotel.ex:{pv_port}",
            f"map http://india.ex http://india.ex:{pv_port}",
        ])

tr = Test.AddTestRun("remap_config reload, succeeds")
tr.Processes.Default.Env = tm.Env
tr.Processes.Default.Command = 'sleep 2; traffic_ctl config reload'

tr = Test.AddTestRun("post update charlie")
await_config_reload = tr.Processes.Process('config_reload_succeeded', 'sleep 30')
await_config_reload.Ready = When.FileContains(tm.Disk.diags_log.Name, "remap.config finished loading", 2)
tr.Processes.Default.StartBefore(await_config_reload)
tr.AddVerifierClientProcess("client_3", replay_file_3, http_ports=[tm.Variables.port])

tr = Test.AddTestRun("post update golf")
tr.AddVerifierClientProcess("client_4", replay_file_4, http_ports=[tm.Variables.port])



'''
Test s3_auth config parsing
'''
#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

ts = Test.MakeATSProcess("ts")
server = Test.MakeOriginServer("server")

Test.testName = "s3_auth: config parsing"

# define the request header and the desired response header
request_header = {
    "headers": "GET /s3-bucket HTTP/1.1\r\nHost: www.example.com\r\n\r\n",
    "timestamp": "1469733493.993",
    "x-amz-security-token":
        "hkMsi6/bfHyBKrSeM/H0hoXeyx8z1yZ/mJ0c+B/TqYx=tTJDjnQWtul38Z9iVJjeH1HB4VT2c=2o3yE3o=I9kmFs/lJDR85qWjB8e5asY/WbjyRpbAzmDipQpboIcYnUYg55bxrQFidV/q8gZa5A9MpR3n=op1C0lWjeBqcEJxpevNZxteSQTQfeGsi98Cdf+On=/SINVlKrNhMnmMsDOLMGx1YYt9d4UsRg1jtVrwxL4Vd/F7aHCZySAXKv+1rkhACR023wpa3dhp+xirGJxSO9LWwvcrTdM4xJo4RS8B40tGENOJ1NKixUJxwN/6og58Oft/u==uleR89Ja=7zszK2H7tX3DqmEYNvNDYQh/7VBRe5otghQtPwJzWpXAGk+Vme4hPPM5K6axH2LxipXzRiIV=oxNs0upKNu1FvuzbCQmkQdKQVmXl0344vngngrgN7wkEfrYtmKwICmpAS0cbW9jdSClgziVo4NaFc/hsIfok=4UA3hVtxIdw74lFNXD0RR7HKXkFPLIn85M7peOZsqMUCfO4gxr7KCfabszQQf0YcP/mt79XK50=WrSJG7oUyn+clUySPhlegqHAfT9a50uSK5WiQmOnGNGLF4wDO10sqKN1xRgQbYHPtwL+Ye0EMisvmYA3==kScorTSGaQWyibSWXAvxq9+IVGBYShVJ6S7DmTT=u/2d/fGEge+Xmbxlftza=cxJ=Md=k1Q71Lp6Boa56d7wtYRpK6tXHJ9I/2r7rN1E4OtwkFqb7SfWV3UXwyUrXyaaNPTIbqnAHnbgUGtuU6pgICpfREiIxVqvKBf6ErbxHRmMmAuYKxk5E9Mn6nnbxR4WTniweKYeDv2w39zge/tss+36Moeuio9d2eoyRFqXhq=rUGtDwX3fzXV0wV+dUojxOYQ57GQDl7+68PwHPcX794OIXuGOxBk83lNIYIcYz3Vc7qnGy6tFTz7f6S9+EZuSGN7TY5VKkT2eWye46DebrDF9Nwzs/FVpTzbPD/KGDIBtFIbazglhKoWe9txqb1QW8vFNNVOEhYa+cViO3g8ZmY1wG960US2zsnX5Eg8Q5a4h3+sxaJSJ4ONiXZWJuAgKRQzcrszu+M5C0ZVoCOv1goEgfNJeSm/yFc/3rx8wmeWLIJFtq65B7zF72HRKq1nthHAguaxXr20nguHpKkDpNBDVa=WwuJsbeGI",
    "body": ""
}

# desired response form the origin server
response_header = {"headers": "HTTP/1.1 200 OK\r\nConnection: close\r\n\r\n", "timestamp": "1469733493.993", "body": "success!"}

# add request/response
server.addResponse("sessionlog.log", request_header, response_header)

ts.Disk.records_config.update(
    {
        'proxy.config.diags.debug.enabled': 1,
        'proxy.config.diags.show_location': 0,
        'proxy.config.diags.debug.tags': 's3_auth',
    })

ts.Setup.CopyAs('rules/v4-parse-test.test_input', Test.RunDirectory)

ts.Disk.remap_config.AddLine(
    f'map http://www.example.com http://127.0.0.1:{server.Variables.Port} \
        @plugin=s3_auth.so \
            @pparam=--config @pparam={Test.RunDirectory}/v4-parse-test.test_input')

# Test Case
tr = Test.AddTestRun()
tr.Processes.Default.Command = f'curl -s -v -H "Host: www.example.com" http://127.0.0.1:{ts.Variables.port}/s3-bucket;'
tr.Processes.Default.ReturnCode = 0
tr.Processes.Default.StartBefore(server)
tr.Processes.Default.StartBefore(ts)
tr.Processes.Default.Streams.stderr.Content = Testers.ContainsExpression("200 OK", "expected 200 response")
tr.Processes.Default.Streams.stderr.Content += Testers.ContainsExpression("Content-Length: 8", "expected content-length 8")
tr.StillRunningAfter = server

ts.Disk.traffic_out.Content = "gold/s3_auth_parsing_ts.gold"
ts.ReturnCode = 0



'''
Test how we handle image/webp
'''
#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

Test.Summary = '''
Checking that we don't serve image/webp to clients that do not support it
'''

Test.ContinueOnFail = True

# Define default ATS
ts = Test.MakeATSProcess("ts")
server = Test.MakeOriginServer("server")

testName = "accept_webp"
request_header = {
    "headers":
        "GET / HTTP/1.1\r\nHost: www.example.com\r\nAccept: image/webp,image/png,image/svg+xml,image/*;q=0.8,video/*;q=0.8,*/*;q=0.5\r\n\r\n",
    "timestamp": "1469733493.993",
    "body": ""
}
response_header = {
    "headers": "HTTP/1.1 200 OK\r\nConnection: close\r\nContent-Type: image/webp\r\nCache-Control: max-age=300\r\n",
    "timestamp": "1469733493.993",
    "body": "xxx"
}
server.addResponse("sessionlog.json", request_header, response_header)

# ATS Configuration
ts.Disk.records_config.update(
    {
        'proxy.config.diags.debug.enabled': 1,
        'proxy.config.diags.debug.tags': 'http_match',
        'proxy.config.http.cache.ignore_accept_mismatch': 0,
        'proxy.config.http.insert_response_via_str': 3,
        'proxy.config.http.cache.http': 1,
        'proxy.config.http.wait_for_cache': 1,
    })

ts.Disk.remap_config.AddLine('map http://www.example.com http://127.0.0.1:{0}'.format(server.Variables.Port))

# Test 1 - Request with image/webp support from the origin
tr = Test.AddTestRun()
tr.Processes.Default.StartBefore(server, ready=When.PortOpen(server.Variables.Port))
tr.Processes.Default.StartBefore(Test.Processes.ts)
tr.Processes.Default.Command = 'curl -s -D - -v --ipv4 --http1.1 -H "Accept: image/webp,image/png,image/svg+xml,image/*;q=0.8,video/*;q=0.8,*/*;q=0.5" -H "Host: www.example.com" http://localhost:{0}/'.format(
    ts.Variables.port)
tr.Processes.Default.ReturnCode = 0
tr.Processes.Default.Streams.stderr = "gold/accept_webp.gold"
tr.StillRunningAfter = ts

# Test 2 - Request with image/webp support from cache
tr = Test.AddTestRun()
tr.Processes.Default.Command = 'curl -s -D - -v --ipv4 --http1.1 -H "Accept: image/webp,image/png,image/svg+xml,image/*;q=0.8,video/*;q=0.8,*/*;q=0.5" -H "Host: www.example.com" http://localhost:{0}/'.format(
    ts.Variables.port)
tr.Processes.Default.ReturnCode = 0
tr.Processes.Default.Streams.stderr = "gold/accept_webp_cache.gold"
tr.StillRunningAfter = ts

# Test 3 - Request without image/webp support going to the origin - NOTE: the origin can't change the content-type :(
tr = Test.AddTestRun()
tr.Processes.Default.Command = 'curl -s -D - -v --ipv4 --http1.1 -H "Accept: image/png,image/svg+xml,image/*;q=0.8,video/*;q=0.8,*/*;q=0.5" -H "Host: www.example.com" http://localhost:{0}/'.format(
    ts.Variables.port)
tr.Processes.Default.ReturnCode = 0
tr.Processes.Default.Streams.stderr = "gold/accept_webp_jpeg.gold"
tr.StillRunningAfter = ts



'''
'''
#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import os

Test.Summary = '''
Test invalid values in range header
'''
Test.ContinueOnFail = True


class InvalidRangeHeaderTest:
    invalidRangeRequestReplayFile = "replays/invalid_range_request.replay.yaml"

    def __init__(self):
        self.setupOriginServer()
        self.setupTS()

    def setupOriginServer(self):
        self.server = Test.MakeVerifierServerProcess("verifier-server1", self.invalidRangeRequestReplayFile)

    def setupTS(self):
        self.ts = Test.MakeATSProcess("ts1")
        self.ts.Disk.records_config.update(
            {
                'proxy.config.diags.debug.enabled': 1,
                'proxy.config.diags.debug.tags': 'http',
                'proxy.config.http.cache.http': 1,
                'proxy.config.http.cache.range.write': 1,
                'proxy.config.http.cache.required_headers': 0,
                'proxy.config.http.insert_age_in_response': 0
            })
        self.ts.Disk.remap_config.AddLine(f"map / http://127.0.0.1:{self.server.Variables.http_port}/",)

    def runTraffic(self):
        tr = Test.AddTestRun()
        tr.AddVerifierClientProcess(
            "client1", self.invalidRangeRequestReplayFile, http_ports=[self.ts.Variables.port], other_args='--thread-limit 1')
        tr.Processes.Default.StartBefore(self.server)
        tr.Processes.Default.StartBefore(self.ts)
        tr.StillRunningAfter = self.server
        tr.StillRunningAfter = self.ts

        # verification
        tr.Processes.Default.Streams.stdout += Testers.ContainsExpression(
            r"Received an HTTP/1 416 response for key 2", "Verify that client receives a 416 response")
        tr.Processes.Default.Streams.stdout += Testers.ContainsExpression(
            r"x-responseheader: failed_response", "Verify that the response came from the server")

    def run(self):
        self.runTraffic()


InvalidRangeHeaderTest().run()



'''
'''
#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import os
import time
from jsonrpc import Request

Test.Summary = '''
regex_revalidate plugin test, MISS (refetch) functionality
'''

# Test description:
# If MISS tag encountered, should load rule as refetch instead of IMS.
# If rule switched from MISS to IMS or vice versa, rule should reset.

Test.SkipUnless(Condition.PluginExists('regex_revalidate.so'), Condition.PluginExists('xdebug.so'))
Test.ContinueOnFail = False

# configure origin server
server = Test.MakeOriginServer("server")

# Define ATS and configure
ts = Test.MakeATSProcess("ts")

Test.testName = "regex_revalidate_miss"
Test.Setup.Copy("metrics_miss.sh")

# default root
request_header_0 = {
    "headers": "GET / HTTP/1.1\r\n" + "Host: www.example.com\r\n" + "\r\n",
    "timestamp": "1469733493.993",
    "body": "",
}

response_header_0 = {
    "headers": "HTTP/1.1 200 OK\r\n" + "Connection: close\r\n" + "Cache-Control: max-age=300\r\n" + "\r\n",
    "timestamp": "1469733493.993",
    "body": "xxx",
}

# cache item path1
request_header_1 = {
    "headers": "GET /path1 HTTP/1.1\r\n" + "Host: www.example.com\r\n" + "\r\n",
    "timestamp": "1469733493.993",
    "body": ""
}
response_header_1 = {
    "headers":
        "HTTP/1.1 200 OK\r\n" + "Connection: close\r\n" + 'Etag: "path1"\r\n' + "Cache-Control: max-age=600,public\r\n" + "\r\n",
    "timestamp": "1469733493.993",
    "body": "abc"
}

server.addResponse("sessionlog.json", request_header_0, response_header_0)
server.addResponse("sessionlog.json", request_header_1, response_header_1)

# Configure ATS server
ts.Disk.plugin_config.AddLine('xdebug.so --enable=x-cache')
ts.Disk.plugin_config.AddLine('regex_revalidate.so -d -c regex_revalidate.conf -l revalidate.log -m reval')

regex_revalidate_conf_path = os.path.join(ts.Variables.CONFIGDIR, 'regex_revalidate.conf')
#curl_and_args = 'curl -s -D - -v -H "x-debug: x-cache" -H "Host: www.example.com"'

path1_rule = 'path1 {}'.format(int(time.time()) + 600)

# Define first revision for when trafficserver starts
ts.Disk.File(regex_revalidate_conf_path, typename="ats:config").AddLine("# Empty")

ts.Disk.remap_config.AddLine('map http://ats/ http://127.0.0.1:{}'.format(server.Variables.Port))

# minimal configuration
ts.Disk.records_config.update(
    {
        'proxy.config.diags.debug.enabled': 1,
        'proxy.config.diags.debug.tags': 'regex_revalidate',
        'proxy.config.http.insert_age_in_response': 0,
        'proxy.config.http.response_via_str': 3,
        'proxy.config.http.cache.http': 1,
        'proxy.config.http.wait_for_cache': 1,
    })

curl_and_args = 'curl -s -D /dev/stdout -o /dev/stderr -x http://127.0.0.1:{}'.format(ts.Variables.port) + ' -H "x-debug: x-cache"'

# 0 Request, cache miss expected
tr = Test.AddTestRun("Cache miss path1")
ps = tr.Processes.Default
ps.StartBefore(server, ready=When.PortOpen(server.Variables.Port))
ps.StartBefore(Test.Processes.ts)
ps.Command = curl_and_args + ' http://ats/path1'
ps.ReturnCode = 0
ps.Streams.stdout.Content = Testers.ContainsExpression("X-Cache: miss", "expected cache miss response")
tr.StillRunningAfter = ts

# 1 Request, cache hit expected
tr = Test.AddTestRun("Cache hit fresh path1")
ps = tr.Processes.Default
ps.Command = curl_and_args + ' http://ats/path1'
ps.ReturnCode = 0
ps.Streams.stdout.Content = Testers.ContainsExpression("X-Cache: hit-fresh", "expected cache hit fresh response")
tr.StillRunningAfter = ts

# 2 Reload, populated config
tr = Test.AddTestRun("Reload config add path1")
ps = tr.Processes.Default
tr.Disk.File(regex_revalidate_conf_path, typename="ats:config").AddLine(path1_rule + ' MISS')
# keep this for debug
tr.Disk.File(regex_revalidate_conf_path + "_tr2", typename="ats:config").AddLine(path1_rule + ' MISS')
tr.StillRunningAfter = ts
tr.StillRunningAfter = server
tr.AddJsonRPCClientRequest(ts, Request.admin_config_reload())
ps.ReturnCode = 0
ps.TimeOut = 5
tr.TimeOut = 5
# Delay it so the reload can catch up the diff between config files timestamps.
tr.DelayStart = 1

# 3 Request, cache miss expected
tr = Test.AddTestRun("Revalidate MISS path1")
ps = tr.Processes.Default
tr.DelayStart = 7
ps.Command = curl_and_args + ' http://ats/path1'
ps.ReturnCode = 0
ps.Streams.stdout.Content = Testers.ContainsExpression("X-Cache: miss", "expected cache miss response")
tr.StillRunningAfter = ts

# 4 Request, cache hit (path1)
tr = Test.AddTestRun("Cache hit fresh path1")
ps = tr.Processes.Default
ps.Command = curl_and_args + ' http://ats/path1'
ps.ReturnCode = 0
ps.Streams.stdout.Content = Testers.ContainsExpression("X-Cache: hit-fresh", "expected cache hit fresh response")
tr.StillRunningAfter = ts

# 5 Reload, change MISS to STALE (resets rule)
tr = Test.AddTestRun("Reload config path1 STALE")
ps = tr.Processes.Default
tr.Disk.File(regex_revalidate_conf_path, typename="ats:config").AddLine(path1_rule + ' STALE')
tr.Disk.File(regex_revalidate_conf_path + "_tr5", typename="ats:config").AddLine(path1_rule + ' STALE')
tr.StillRunningAfter = ts
tr.StillRunningAfter = server
tr.AddJsonRPCClientRequest(ts, Request.admin_config_reload())
ps.ReturnCode = 0
ps.TimeOut = 5
tr.TimeOut = 5
# Delay it so the reload can catch up the diff between config files timestamps.
tr.DelayStart = 1

# 6 Request, cache stale expected
tr = Test.AddTestRun("Cache stale path1")
ps = tr.Processes.Default
tr.DelayStart = 7
ps.Command = curl_and_args + ' http://ats/path1'
ps.ReturnCode = 0
ps.Streams.stdout.Content = Testers.ContainsExpression("X-Cache: hit-stale", "expected cache hit stale response")
tr.StillRunningAfter = ts

# 7 Reload, change STALE to MISS (resets rule)
tr = Test.AddTestRun("Reload config path1 MISS")
ps = tr.Processes.Default
tr.Disk.File(regex_revalidate_conf_path, typename="ats:config").AddLine(path1_rule + ' MISS')
tr.Disk.File(regex_revalidate_conf_path + "_tr7", typename="ats:config").AddLine(path1_rule + ' MISS')
tr.StillRunningAfter = ts
tr.StillRunningAfter = server
tr.AddJsonRPCClientRequest(ts, Request.admin_config_reload())
ps.ReturnCode = 0
ps.TimeOut = 5
tr.TimeOut = 5
# Delay it so the reload can catch up the diff between config files timestamps.
tr.DelayStart = 1

# 8 Request, cache miss expected
tr = Test.AddTestRun("Cache mis path1")
ps = tr.Processes.Default
tr.DelayStart = 7
ps.Command = curl_and_args + ' http://ats/path1'
ps.ReturnCode = 0
ps.Streams.stdout.Content = Testers.ContainsExpression("X-Cache: miss", "expected cache miss response")
tr.StillRunningAfter = ts

# 9 Request, cache hit expected
tr = Test.AddTestRun("Cache hit path1")
ps = tr.Processes.Default
tr.DelayStart = 5
ps.Command = curl_and_args + ' http://ats/path1'
ps.ReturnCode = 0
ps.Streams.stdout.Content = Testers.ContainsExpression("X-Cache: hit-fresh", "expected cache hit response")
tr.StillRunningAfter = ts

# 10 Reload, no changes
tr = Test.AddTestRun("Reload no changes")
ps = tr.Processes.Default
tr.StillRunningAfter = ts
tr.StillRunningAfter = server
tr.AddJsonRPCClientRequest(ts, Request.admin_config_reload())
ps.ReturnCode = 0
ps.TimeOut = 5
tr.TimeOut = 5
# Delay it so the reload can catch up the diff between config files timestamps.
tr.DelayStart = 1

# 11 Request again, cache hit expected
tr = Test.AddTestRun("Cache hit path1")
ps = tr.Processes.Default
tr.DelayStart = 7
ps.Command = curl_and_args + ' http://ats/path1'
ps.ReturnCode = 0
ps.Streams.stdout.Content = Testers.ContainsExpression("X-Cache: hit-fresh", "expected cache hit response")
tr.StillRunningAfter = ts

# 12 Stage - Reload of same time updated rules file
tr = Test.AddTestRun("Reload same config")
ps = tr.Processes.Default
tr.StillRunningAfter = ts
tr.StillRunningAfter = server
ps.Command = f'touch {regex_revalidate_conf_path} ; traffic_ctl config reload'
ps.Env = ts.Env
ps.ReturnCode = 0
ps.TimeOut = 5
tr.TimeOut = 5

# 13 Request, Cache hit expected
tr = Test.AddTestRun("Cache hit path1")
ps = tr.Processes.Default
tr.DelayStart = 5
ps.Command = curl_and_args + ' http://ats/path1'
ps.ReturnCode = 0
ps.Streams.stdout.Content = Testers.ContainsExpression("X-Cache: hit-fresh", "expected cache hit response")
tr.StillRunningAfter = ts

# 14 Stats check
tr = Test.AddTestRun("Check stats")
tr.DelayStart = 5
tr.Processes.Default.Command = f"bash -c ./metrics_miss.sh"
tr.Processes.Default.Env = ts.Env
tr.Processes.Default.ReturnCode = 0
tr.StillRunningAfter = ts



'''Implement a client that sends many small DATA frames.'''

#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import argparse
from email.message import EmailMessage as HttpHeaders
import logging
import math
import socket
import sys
import ssl
from OpenSSL.SSL import Error as SSLError
from OpenSSL.SSL import SysCallError as SSLSysCallError
import statistics
import traceback
import time

from typing import Dict, List, Tuple

import eventlet
from h2.connection import H2Connection
from h2.events import StreamEnded, ResponseReceived, DataReceived, TrailersReceived, StreamReset, ConnectionTerminated
from h2.exceptions import StreamClosedError


def get_body_text() -> bytes:
    """Create a body of text to send in the request."""
    body_items: List[str] = []
    for i in range(100):
        # Create a chunk of 0 padded bytes, followed by a space.
        chunk_payload = f'{i:06x} '.encode("utf-8")
        body_items.append(chunk_payload)
    return b''.join(body_items)


class RequestInfo:
    """POD for request headers, etc."""

    def __init__(self, stream_id: int, headers: Dict[str, str], body: str):
        self.stream_id: int = stream_id
        self.headers: Dict[str, str] = headers
        self.body_bytes: str = body


class ResponseInfo:
    """POD for response headers, etc."""

    def __init__(
            self,
            status: int,
            headers: Dict[bytes, bytes],
            body: bytes,
            trailers: Dict[bytes, bytes] = None,
            errors: List[str] = None):
        self.status_code: int = status
        self.headers: Dict[bytes, bytes] = headers
        self.body_bytes: bytes = body
        self.trailers: Dict[bytes, bytes] = trailers
        self.errors: List[str] = errors


def print_transaction(request: RequestInfo, response: ResponseInfo) -> None:
    """Print a description of the transaction.

    :param request: The details about the request.
    :param response: The details about the response.
    """

    description = "\n==== REQUEST HEADERS ====\n"
    for k, v in request.headers.items():
        if isinstance(k, bytes):
            k, v = (k.decode('ascii'), v.decode('ascii'))
        description += f"{k}: {v}\n"

    if request.body_bytes is not None:
        description += f"\n==== REQUEST BODY ====\n{request.body_bytes}\n"

    description += "\n==== RESPONSE ====\n"
    description += f"{response.status_code}\n"

    description += "\n==== RESPONSE HEADERS ====\n"
    for k, v in response.headers:
        if isinstance(k, bytes):
            k, v = (k.decode('ascii'), v.decode('ascii'))
        description += f"{k}: {v}\n"

    if response.body_bytes is not None:
        description += f"\n==== RESPONSE BODY ====\n{response.body_bytes.decode()}\n"

    if response.trailers is not None:
        description += "\n==== RESPONSE TRAILERS ====\n"
        for k, v in response.trailers.items():
            if isinstance(k, bytes):
                k, v = (k.decode('ascii'), v.decode('ascii'))
            description += "{k}: {v}\n"

    description += "\n==== END ====\n"

    logging.info(description)


class Http2Connection:
    '''
    This class manages a single HTTP/2 connection to a server. It is not
    thread-safe. For our purpose though, no lock is neccessary as the streams of
    each connection are processed sequentially.
    '''

    def __init__(self, sock, h2conn):
        self.sock = sock
        self.conn = h2conn

    def send_request(self, request: RequestInfo) -> Tuple[ResponseInfo, List[int]]:
        '''
        Sends a request to the h2 connection and returns the response object containing the headers, body, and possible errors.
        '''
        self.conn.send_headers(request.stream_id, request.headers.items())
        logging.info(f'Sent headers.')
        # Send the data over the socket.
        self.sock.sendall(self.conn.data_to_send())
        response_headers_raw = None
        response_body = b''
        response_stream_ended = False
        request_stream_ended = False
        trailers = None
        errors = []
        bytes_sent = 0
        bytes_left = len(request.body_bytes)
        data_frame_differentials: List[int] = []
        time_of_last_frame = time.perf_counter_ns()
        while not response_stream_ended:

            send_window = self.conn.local_flow_control_window(request.stream_id)
            bytes_to_send = min(send_window, bytes_left)
            # Send one byte at a time, every millisecond.
            while bytes_to_send > 0:
                chunk_size = 1
                byte_to_send = request.body_bytes[bytes_sent:bytes_sent + chunk_size]
                logging.debug(f'Sending {byte_to_send}')
                self.conn.send_data(request.stream_id, byte_to_send)
                self.sock.sendall(self.conn.data_to_send())
                bytes_left -= chunk_size
                bytes_sent += chunk_size
                bytes_to_send -= chunk_size
                time.sleep(0.001)

            if not request_stream_ended and bytes_left == 0:
                logging.debug('Closing the connection')
                self.conn.end_stream(request.stream_id)
                request_stream_ended = True

            logging.debug('Reading any responses from the socket')
            data = self.sock.recv(65536 * 1024)
            if not data:
                break

            # Feed raw data into h2 engine, and process resulting events.
            logging.debug('Feeding the data into the connection')
            events = self.conn.receive_data(data)
            have_counted_data_delay = False
            for event in events:
                if isinstance(event, ResponseReceived):
                    # Received response headers.
                    response_headers_raw = event.headers
                    time_of_last_frame = time.perf_counter_ns()
                    logging.info('Received response headers.')
                if isinstance(event, DataReceived):
                    # Update flow control so the server doesn't starve us.
                    self.conn.acknowledge_received_data(event.flow_controlled_length, event.stream_id)
                    # Received more response body data.
                    response_body += event.data
                    current_time = time.perf_counter_ns()
                    ms_delay = (current_time - time_of_last_frame) / (1000 * 1000)
                    if not have_counted_data_delay:
                        data_frame_differentials.append(ms_delay)
                        time_of_last_frame = time.perf_counter_ns()
                        logging.debug(f"Received {len(event.data)} bytes of data after {ms_delay} ms")
                        have_counted_data_delay = True
                if isinstance(event, TrailersReceived):
                    # Received trailer headers.
                    trailers = event.headers
                if isinstance(event, StreamReset):
                    # Stream reset by the server.
                    logging.debug(f"Received RST_STREAM from the server: {event}")
                    errors.append('StreamReset')
                    response_stream_ended = True
                    break
                if isinstance(event, ConnectionTerminated):
                    # Received GOAWAY frame from the server.
                    logging.debug(f"Received GOAWAY from the server: {event}")
                    errors.append('ConnectionTerminated')
                    response_stream_ended = True
                    break
                if isinstance(event, StreamEnded):
                    # Received complete response body.
                    logging.info('Received stream end.')
                    response_stream_ended = True
                    break

            if not errors:
                # Send any pending data to the server.
                self.sock.sendall(self.conn.data_to_send())

        # Decode the header fields.
        status_code = next((t[1] for t in response_headers_raw if t[0].decode() == ':status'), None)
        status_code = int(status_code) if status_code else 0
        return ResponseInfo(status_code, response_headers_raw, response_body, trailers, errors), data_frame_differentials

    def close(self):
        """Tell the server we are closing the h2 connection."""
        self.conn.close_connection()
        self.sock.sendall(self.conn.data_to_send())
        self.sock.close()


def create_ssl_context(cert):
    """
    Create a SSL context with the given cert file.
    """
    ctx = ssl.create_default_context()
    ctx.set_alpn_protocols(['h2', 'http/1.1'])
    # Load the cert file
    ctx.load_cert_chain(cert)
    # Do not verify the server's certificate
    ctx.check_hostname = False
    ctx.verify_mode = ssl.CERT_NONE
    return ctx


def send_http2_request_to_server(hostname: str, port: int, cert_file: str, write_timeout: int) -> int:
    """Establish a connection with the server and send a request.

    :param hostname: The hostname to use for the :authority header.
    :param port: The port to connect to.
    :param cert_file: The TLS certificate file.
    :param write_timeout: The expected maximum amount of time frames should be delivered.

    :return: 0 if the request was successful, 1 otherwise.
    """

    request_headers = HttpHeaders()
    request_headers.add_header(':method', 'GET')
    request_headers.add_header(':path', '/some/path')
    request_headers.add_header(':authority', hostname)
    request_headers.add_header(':scheme', 'https')

    scheme = request_headers[':scheme']
    replay_server = f"127.0.0.1:{port}"
    path = request_headers[':path']
    authority = request_headers.get(':authority', '')

    stream_id = 1
    body = get_body_text()
    request: RequestInfo = RequestInfo(stream_id, request_headers, body)

    try:
        # Open a socket to the server and initiate TLS/SSL.
        ssl_context = create_ssl_context(cert=cert_file)
        setattr(ssl_context, "old_wrap_socket", ssl_context.wrap_socket)

        def new_wrap_socket(sock, *args, **kwargs):
            # Make the SNI line up with the :authority header value.
            kwargs['server_hostname'] = hostname
            return ssl_context.old_wrap_socket(sock, *args, **kwargs)

        setattr(ssl_context, "wrap_socket", new_wrap_socket)
        # Opens a connection to the server.
        logging.info(f"Connecting to '{scheme}://{replay_server}' with request to '{authority}{path}'")
        sock = socket.create_connection(('127.0.0.1', port))
        sock = ssl_context.wrap_socket(sock)

        # Initiate an HTTP/2 connection.
        http2_connection = H2Connection()
        http2_connection.initiate_connection()
        # Initial SETTINGS frame, etc.
        sock.sendall(http2_connection.data_to_send())
        client = Http2Connection(sock, http2_connection)
        response, data_delays = client.send_request(request)
        if response.errors:
            try:
                if 'StreamReset' in response.errors:
                    http2_connection.reset_stream(stream_id)
                if 'ConnectionTerminated' in response.errors:
                    http2_connection.close_connection(last_stream_id=0)
            except StreamClosedError as err:
                logging.error(err)
            return 1
        else:
            client.close()
    except Exception as e:
        logging.error(f"Connection to '{replay_server}' initiated with request to "
                      f"'{scheme}://{authority}{path}' failed: {e}")
        traceback.print_exc(file=sys.stdout)
        return 1

    print_transaction(request, response)
    logging.info(f'Smallest delay: {min(data_delays)} ms')
    logging.info(f'Largest delay: {max(data_delays)} ms')
    average = statistics.mean(data_delays)
    logging.info(f'Average delay over {len(data_delays)} reads: {average} ms')
    isclose = math.isclose(average, write_timeout, rel_tol=0.2)
    if isclose:
        logging.info(f'Average delay of {average} is within 20% of the expected delay: {write_timeout} ms')
    else:
        logging.info(f'Average delay of {average} is not within 20% of the expected delay: {write_timeout} ms')

    return 0 if isclose else 1


def parse_args() -> argparse.Namespace:
    """Parse the command line arguments.

    :return: The parsed arguments.
    """
    parser = argparse.ArgumentParser(description=__doc__)
    parser.add_argument('hostname', type=str, help='The hostname to use for the :authority header.')
    parser.add_argument('port', type=int, help='The port to connect to.')
    parser.add_argument('cert', type=str, help='The TLS certificate file.')
    parser.add_argument('write_timeout', type=int, help='The expected maximum amount of time frames should be delivered.')
    parser.add_argument('-v', '--verbose', action='store_true', help='Enable verbose logging.')
    return parser.parse_args()


def main() -> int:
    """Start the client."""
    args = parse_args()

    log_level = logging.DEBUG if args.verbose else logging.INFO
    logging.basicConfig(level=log_level, format='%(asctime)s - %(levelname)s - %(message)s')

    return send_http2_request_to_server(args.hostname, args.port, args.cert, args.write_timeout)


if __name__ == '__main__':
    sys.exit(main())



'''
'''
#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

Test.Summary = '''
cache_range_requests with cachekey as a global plugin
'''

# Test description:
# Preload the cache with the entire asset to be range requested.
# Reload remap rule with cache_range_requests plugin
# Request content through the cache_range_requests plugin

Test.SkipUnless(
    Condition.PluginExists('cache_range_requests.so'),
    Condition.PluginExists('cachekey.so'),
    Condition.PluginExists('xdebug.so'),
)
Test.ContinueOnFail = False
Test.testName = "cache_range_requests_cachekey_global"

# Define and configure ATS, enable traffic_ctl config reload
ts = Test.MakeATSProcess("ts")

# Define and configure origin server
server = Test.MakeOriginServer("server", lookup_key="{%uuid}")

# default root
req_chk = {
    "headers": "GET / HTTP/1.1\r\n" + "Host: www.example.com\r\n" + "uuid: none\r\n" + "\r\n",
    "timestamp": "1469733493.993",
    "body": ""
}

res_chk = {"headers": "HTTP/1.1 200 OK\r\n" + "Connection: close\r\n" + "\r\n", "timestamp": "1469733493.993", "body": ""}

server.addResponse("sessionlog.json", req_chk, res_chk)

body = "lets go surfin now"
bodylen = len(body)

# this request should work
req_full = {
    "headers": "GET /path HTTP/1.1\r\n" + "Host: www.example.com\r\n" + "Accept: */*\r\n" + "uuid: full\r\n" + "\r\n",
    "timestamp": "1469733493.993",
    "body": ""
}

res_full = {
    "headers":
        "HTTP/1.1 206 Partial Content\r\n" + "Accept-Ranges: bytes\r\n" + 'Etag: "foo"\r\n' +
        "Cache-Control: public, max-age=500\r\n" + "Connection: close\r\n" + "\r\n",
    "timestamp": "1469733493.993",
    "body": body
}

server.addResponse("sessionlog.json", req_full, res_full)

# this request should work
req_good = {
    "headers":
        "GET /path HTTP/1.1\r\n" + "Host: www.example.com\r\n" + "Accept: */*\r\n" + "Range: bytes=0-\r\n" +
        "uuid: range_full\r\n" + "\r\n",
    "timestamp": "1469733493.993",
    "body": ""
}

res_good = {
    "headers":
        "HTTP/1.1 206 Partial Content\r\n" + "Accept-Ranges: bytes\r\n" + 'Etag: "foo"\r\n' +
        "Cache-Control: public, max-age=500\r\n" + "Content-Range: bytes 0-{0}/{0}\r\n".format(bodylen) + "Connection: close\r\n" +
        "\r\n",
    "timestamp": "1469733493.993",
    "body": body
}

server.addResponse("sessionlog.json", req_good, res_good)

# this request should fail with a cache_range_requests asset
req_fail = {
    "headers":
        "GET /path HTTP/1.1\r\n" + "Host: www.fail.com\r\n" + "Accept: */*\r\n" + "Range: bytes=0-\r\n" + "uuid: range_fail\r\n" +
        "\r\n",
    "timestamp": "1469733493.993",
    "body": ""
}

res_fail = {
    "headers":
        "HTTP/1.1 206 Partial Content\r\n" + "Accept-Ranges: bytes\r\n" + 'Etag: "foo"\r\n' +
        "Cache-Control: public, max-age=500\r\n" + "Content-Range: bytes 0-{0}/{0}\r\n".format(bodylen) + "Connection: close\r\n" +
        "\r\n",
    "timestamp": "1469733493.993",
    "body": body
}

server.addResponse("sessionlog.json", req_fail, res_fail)

# cache range requests plugin remap, working config
ts.Disk.remap_config.AddLine('map http://www.example.com http://127.0.0.1:{}'.format(server.Variables.Port))

# improperly configured cache_range_requests with cachekey
ts.Disk.remap_config.AddLine('map http://www.fail.com http://127.0.0.1:{}'.format(server.Variables.Port))

# cache debug
ts.Disk.plugin_config.AddLines(
    [
        'cachekey.so --include-headers=Range --static-prefix=foo',
        'cache_range_requests.so --no-modify-cachekey',
        'xdebug.so --enable=x-cache,x-cache-key,x-parentselection-key',
    ])

# minimal configuration
ts.Disk.records_config.update(
    {
        'proxy.config.diags.debug.enabled': 1,
        'proxy.config.diags.debug.tags': 'cachekey|cache_range_requests',
    })

curl_and_args = 'curl -s -D /dev/stdout -o /dev/stderr -x localhost:{} -H "x-debug: x-cache-key"'.format(ts.Variables.port)

# 0 Test - Fetch full asset via range
tr = Test.AddTestRun("asset fetch via range")
ps = tr.Processes.Default
ps.StartBefore(server, ready=When.PortOpen(server.Variables.Port))
ps.StartBefore(Test.Processes.ts)
ps.Command = curl_and_args + ' http://www.example.com/path -r0- -H "uuid: full"'
ps.ReturnCode = 0
ps.Streams.stdout.Content = Testers.ContainsExpression(
    "X-Cache-Key: /foo/Range:bytes=0-/path", "expected cachekey style range request in cachekey")
tr.StillRunningAfter = ts



'''
Verify HTTP body buffering.
'''
#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import os

Test.SkipUnless(Condition.PluginExists('request_buffer.so'))


def int_to_hex_string(int_value):
    '''
    Convert the given int value to a hex string with no '0x' prefix.

    >>> int_to_hex_string(0)
    '0'
    >>> int_to_hex_string(1)
    '1'
    >>> int_to_hex_string(10)
    'a'
    >>> int_to_hex_string(16)
    '10'
    >>> int_to_hex_string(17)
    'f1'
    '''
    if not isinstance(int_value, int):
        raise ValueError("Input should be an int type.")
    return hex(int_value).split('x')[1]


class BodyBufferTest:

    def __init__(cls, description):
        Test.Summary = description
        cls._origin_max_connections = 3
        cls.setupOriginServer()
        cls.setupTS()

    def setupOriginServer(self):
        self._server = Test.MakeOriginServer("server")
        self.content_length_request_body = "content-length request"
        self.content_length_size = len(self.content_length_request_body)
        request_header = {
            "headers":
                "POST /contentlength HTTP/1.1\r\n"
                "Host: www.example.com\r\n"
                f"Content-Length: {self.content_length_size}\r\n\r\n",
            "timestamp": "1469733493.993",
            "body": self.content_length_request_body
        }
        content_length_response_body = "content-length response"
        content_length_response_size = len(content_length_response_body)
        response_header = {
            "headers":
                "HTTP/1.1 200 OK\r\n"
                "Server: microserver\r\n"
                f"Content-Length: {content_length_response_size}\r\n\r\n"
                "Connection: close\r\n\r\n",
            "timestamp": "1469733493.993",
            "body": content_length_response_body
        }
        self._server.addResponse("sessionlog.json", request_header, response_header)

        self.chunked_request_body = "chunked request"
        hex_size = int_to_hex_string(len(self.chunked_request_body))
        self.encoded_chunked_request = f"{hex_size}\r\n{self.chunked_request_body}\r\n0\r\n\r\n"
        self.encoded_chunked_size = len(self.content_length_request_body)
        request_header2 = {
            "headers":
                "POST /chunked HTTP/1.1\r\n"
                "Transfer-Encoding: chunked\r\n"
                "Host: www.example.com\r\n"
                "Connection: keep-alive\r\n\r\n",
            "timestamp": "1469733493.993",
            "body": self.encoded_chunked_request
        }
        self.chunked_response_body = "chunked response"
        hex_size = int_to_hex_string(len(self.chunked_response_body))
        self.encoded_chunked_response = f"{hex_size}\r\n{self.chunked_response_body}\r\n0\r\n\r\n"
        response_header2 = {
            "headers": "HTTP/1.1 200 OK\r\n"
                       "Transfer-Encoding: chunked\r\n"
                       "Server: microserver\r\n"
                       "Connection: close\r\n\r\n",
            "timestamp": "1469733493.993",
            "body": self.encoded_chunked_response
        }
        self._server.addResponse("sessionlog.json", request_header2, response_header2)

    def setupTS(self):
        self._ts = Test.MakeATSProcess("ts", select_ports=False)
        self._ts.Disk.remap_config.AddLine(f'map / http://127.0.0.1:{self._server.Variables.Port}')
        Test.PrepareInstalledPlugin('request_buffer.so', self._ts)
        self._ts.Disk.records_config.update(
            {
                'proxy.config.diags.debug.enabled': 1,
                'proxy.config.diags.debug.tags': 'request_buffer',
            })

        self._ts.Disk.traffic_out.Content = Testers.ContainsExpression(
            rf"request_buffer_plugin gets the request body with length\[{self.content_length_size}\]",
            "Verify that the plugin parsed the content-length request body data.")
        self._ts.Disk.traffic_out.Content += Testers.ContainsExpression(
            rf"request_buffer_plugin gets the request body with length\[{self.encoded_chunked_size}\]",
            "Verify that the plugin parsed the chunked request body.")

    def run(self):
        tr = Test.AddTestRun()
        # Send both a Content-Length request and a chunked-encoded request.
        tr.Processes.Default.Command = (
            f'curl -v http://127.0.0.1:{self._ts.Variables.port}/contentlength -d "{self.content_length_request_body}" --next '
            f'-v http://127.0.0.1:{self._ts.Variables.port}/chunked -H "Transfer-Encoding: chunked" -d "{self.chunked_request_body}"'
        )
        tr.Processes.Default.ReturnCode = 0
        tr.Processes.Default.StartBefore(self._server)
        tr.Processes.Default.StartBefore(Test.Processes.ts)
        tr.Processes.Default.Streams.stderr = "200.gold"


bodyBufferTest = BodyBufferTest("Test request body buffering.")
bodyBufferTest.run()



#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

Test.Summary = '''
Test reloading ssl_multicert.config with errors and keeping around the old ssl config structure
'''

sni_domain = 'example.com'

ts = Test.MakeATSProcess("ts", enable_tls=True)
server = Test.MakeOriginServer("server")
server2 = Test.MakeOriginServer("server3")
request_header = {"headers": f"GET / HTTP/1.1\r\nHost: {sni_domain}\r\n\r\n", "timestamp": "1469733493.993", "body": ""}

response_header = {"headers": "HTTP/1.1 200 OK\r\nConnection: close\r\n\r\n", "timestamp": "1469733493.993", "body": ""}
server.addResponse("sessionlog.json", request_header, response_header)

ts.Disk.records_config.update(
    {
        'proxy.config.ssl.server.cert.path': f'{ts.Variables.SSLDir}',
        'proxy.config.ssl.server.private_key.path': f'{ts.Variables.SSLDir}',
        'proxy.config.ssl.server.multicert.exit_on_load_fail': 0,
    })

ts.addDefaultSSLFiles()

ts.Disk.remap_config.AddLine(f'map / http://127.0.0.1:{server.Variables.Port}')

ts.Disk.ssl_multicert_config.AddLine('dest_ip=* ssl_cert_name=server.pem ssl_key_name=server.key')

tr = Test.AddTestRun("ensure we can connect for SNI $sni_domain")
tr.Processes.Default.StartBefore(Test.Processes.ts)
tr.Processes.Default.StartBefore(server)
tr.StillRunningAfter = ts
tr.StillRunningAfter = server
tr.Processes.Default.Command = f"curl -q -s -v -k --resolve '{sni_domain}:{ts.Variables.ssl_port}:127.0.0.1' https://{sni_domain}:{ts.Variables.ssl_port}"
tr.Processes.Default.ReturnCode = 0
tr.Processes.Default.Streams.stdout = Testers.ExcludesExpression("Could Not Connect", "Check response")
tr.Processes.Default.Streams.stderr = Testers.IncludesExpression(f"CN={sni_domain}", "Check response")

tr2 = Test.AddTestRun("Update config files")
# Update the configs
sslcertpath = ts.Disk.ssl_multicert_config.AbsPath

tr2.Disk.File(sslcertpath, id="ssl_multicert_config", typename="ats:config")
tr2.Disk.ssl_multicert_config.AddLines(
    [
        'ssl_cert_name=server_does_not_exist.pem ssl_key_name=server_does_not_exist.key',
        'dest_ip=* ssl_cert_name=server.pem_doesnotexist ssl_key_name=server.key',
    ])
tr2.StillRunningAfter = ts
tr2.StillRunningAfter = server
tr2.Processes.Default.Command = 'echo Updated configs'
tr2.Processes.Default.Env = ts.Env
tr2.Processes.Default.ReturnCode = 0

tr2reload = Test.AddTestRun("Reload config")
tr2reload.StillRunningAfter = ts
tr2reload.StillRunningAfter = server
tr2reload.Processes.Default.Command = 'traffic_ctl config reload'
tr2reload.Processes.Default.Env = ts.Env
tr2reload.Processes.Default.ReturnCode = 0
ts.Disk.diags_log.Content = Testers.ContainsExpression('ERROR: ', 'ERROR')

# Reload of ssl_multicert.config should fail, BUT the old config structure
# should be in place to successfully answer for the test domain
tr3 = Test.AddTestRun("Make request again for $sni_domain")
# Wait for the reload to complete
tr3.Processes.Default.StartBefore(server2, ready=When.FileContains(ts.Disk.diags_log.Name, 'failed to load certificate ', 1))
tr3.StillRunningAfter = ts
tr3.StillRunningAfter = server
tr3.Processes.Default.Command = f"curl -q -s -v -k --resolve '{sni_domain}:{ts.Variables.ssl_port}:127.0.0.1' https://{sni_domain}:{ts.Variables.ssl_port}"
tr3.Processes.Default.ReturnCode = 0
tr3.Processes.Default.Streams.stdout = Testers.ExcludesExpression("Could Not Connect", "Check response")
tr3.Processes.Default.Streams.stderr = Testers.IncludesExpression(f"CN={sni_domain}", "Check response")

##########################################################################
# Ensure ATS fails/exits when non-existent cert is specified
# Also, not explicitly setting proxy.config.ssl.server.multicert.exit_on_load_fail
# to catch if the current default (1) changes in the future

ts2 = Test.MakeATSProcess("ts2", enable_tls=True)
ts2.Disk.ssl_multicert_config.AddLines([
    'dest_ip=* ssl_cert_name=server.pem_doesnotexist ssl_key_name=server.key',
])

tr4 = Test.AddTestRun()
tr4.Processes.Default.Command = 'echo Waiting'
tr4.Processes.Default.ReturnCode = 0
tr4.Processes.Default.StartBefore(ts2)

ts2.ReturnCode = 70  # ink_fatal will exit with EX_SOFTWARE.
ts2.Ready = 0  # Need this to be 0 because we are testing shutdown, this is to make autest not think ats went away for a bad reason.
ts2.Disk.traffic_out.Content = Testers.ExcludesExpression(
    'Traffic Server is fully initialized', 'process should fail when invalid certificate specified')
ts2.Disk.diags_log.Content = Testers.IncludesExpression('FATAL: failed to load SSL certificate file', 'check diags.log"')
